import dataclasses
import datetime
import pandas as pd
import numpy as np
import math
import statsmodels.api as sm
from typing import Optional
from statsmodels import regression
from scipy.stats import gmean
from functools import partial
import statsmodels.api as sm
from statsmodels.tsa.ar_model import AutoReg
from scipy.stats import kurtosis, skew
from decimal import Decimal, ROUND_HALF_UP, Context
from pandas.tseries.offsets import DateOffset

class CalculatorBase:

    TRADING_DAYS_PER_YEAR = 242
    TOTAL_DAYS_PER_YEAR = 365
    TOTAL_WEEKS_PER_YEAR = 52
    TOTAL_MONTHS_PER_YEAR = 12
    STOCK_IPO_NUM_MARKET = {
        '深交所中小板': 500,
        '深交所创业板': 500,
        '上交所主板': 1000,
        '上交所科创板': 500,
        '深交所主板': 500,
    }

    @staticmethod
    def get_begin_date_dic(week_end_date):
        begin_date_period_dic = {
            'w1': DateOffset(weeks=1),
            'm1': DateOffset(months=1),
            'm3': DateOffset(months=3),
            'm6': DateOffset(months=6),
            'y1': DateOffset(years=1),
            'y2': DateOffset(years=2),
            'y3': DateOffset(years=3),
            'y4': DateOffset(years=4),
            'y5': DateOffset(years=5),
        }
        last_date = datetime.datetime.now().date()
        begin_date_dic = {}
        for name, value in begin_date_period_dic.items():
            if name == 'w1':
                bd = (week_end_date - value).date()
            else:
                bd = (week_end_date - value).date()
            begin_date_dic[name] = bd
        return last_date, begin_date_dic

    @staticmethod
    def calc_cl_alpha_beta(total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'beta_bull':np.nan,
                    'beta_bear':np.nan,
                    
                    }
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                'beta_bull':est2.params[0],
                'beta_bear':-est2.params[1],
                }
    
    @staticmethod
    def calc_hm_alpha_beta(total: np.ndarray):
        # 牛市追踪 市场上涨是alpha
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 1][X[:, 1] < 0] = 0
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                }

    @staticmethod
    def stat_round_pct(x):
        s = str(round(100*x,2))
        s = s[::-1].zfill(4)[::-1]
        return s + '%'

    @staticmethod
    def calc_continue_regress_v(monthly_ret, risk_free_rate):
        # 半年周期
        window = 6
        yearly_multiplier = 2
        period_num = 12
        annual_ret = monthly_ret.rolling(window=window).apply(lambda x : x.sum() * yearly_multiplier - risk_free_rate)
        annual_vol = monthly_ret.rolling(window=window).apply(lambda x : x.std(ddof=1) * np.sqrt(period_num))
        sharpe = (annual_ret / annual_vol)
        dic = {np.Inf:np.nan,-np.Inf:np.nan}
        sharpe = sharpe.replace(dic).dropna()
        if sharpe.shape[0] < 6:
            return np.nan
        mod = AutoReg(endog=sharpe.values, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    @staticmethod
    def calc_modify_annual_vol(sr: pd.core.series.Series):        
        import json
        import requests
        url = 'http://69.231.155.101:8006/api/v1/basic/trade_dates'
        response = requests.get(url)    
        dts = json.loads(response._content)['data']['datetime']
        dts = pd.to_datetime(dts)
        dts = np.array([i.date() for i in dts])
        dts = pd.Series(data=1,index=dts)
        sr_ret = sr.pct_change(1)
        sr_ret.name = 'ret'
        sr_ret.index.name = 'this_day'
        sr_ret = sr_ret.reset_index()
        sr_ret.loc[:,'last_day'] = sr_ret.this_day.shift(1)
        sr_ret = sr_ret.dropna()
        sr_ret.loc[:,'trade_dates'] = sr_ret.apply(lambda x: dts.loc[x.last_day:x.this_day].shape[0], axis=1)
        sr_ret.loc[:,'modify_ret'] = sr_ret.ret / np.sqrt(sr_ret.trade_dates)
        annual_vol = sr_ret.modify_ret.replace({np.inf:np.nan,-np.inf:np.nan}).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
        return annual_vol

    @staticmethod
    def calc_recent_ret(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        frequency:         str='1D'):
        try:
            assert frequency in ['1D'], f'frequency provided: {frequency} is not included, must in 1D'
            dates = pd.to_datetime(dates)
            dates = [i.date() for i in dates]
            _end_dt = dates[-1]
            last_date, begin_date_dic = CalculatorBase.get_begin_date_dic(week_end_date = _end_dt)
            assert len(dates) == len(values), 'date and port_values has different numbers'
            sr = pd.Series(values, index=dates).sort_index().dropna()
            sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
            sr.index = [i.date() for i in sr.index]
            _begin_date = sr.index[0]
            _natural_year_begin_date = datetime.date(last_date.year,1,1)
            _sr = sr.loc[:_natural_year_begin_date]
            if _sr.shape[0] > 2:
                _natural_year_begin_date = _sr.index[-2] if _natural_year_begin_date in sr.index else _sr.index[-1]
                recent_natural_year_ret = sr.loc[:last_date][-1] / sr.loc[_natural_year_begin_date] - 1
            else:
                _sr = sr.loc[_natural_year_begin_date:]
                if _sr.empty:
                    recent_natural_year_ret = 0
                else:
                    _natural_year_begin_date = _sr.index[0]
                    recent_natural_year_ret = sr.loc[:last_date][-1] / sr.loc[_natural_year_begin_date] - 1
            _res = {}
            for _period, _period_begin_date in begin_date_dic.items():
                if _period_begin_date >= _begin_date and sr.index[-1] > _period_begin_date:
                    b_d = sr.loc[_period_begin_date:].index[0]
                    e_d = sr.loc[:last_date].index[-1]
                    if e_d == b_d:
                        _res[_period] = 0
                        _res[f'{_period}_annual'] = 0
                        _res[f'{_period}_mdd'] = 0
                    else:
                        sr_sub = sr.loc[b_d:e_d]
                        days = (e_d - b_d).days
                        last_unit_nav = sr_sub.iloc[-1] / sr_sub.iloc[0]
                        _res[_period] = last_unit_nav - 1
                        _res[f'{_period}_annual'] = math.pow(last_unit_nav, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1
                        _res[f'{_period}_mdd'] = 1 - (sr_sub / sr_sub.cummax()).min()
                    
                    if _period == 'w1':
                        _res['w1_begin_date'] = sr.loc[_period_begin_date:].index[0]
                        _res['w1_end_date'] = sr.loc[:last_date].index[-1]
                else:
                    _res[_period] = np.nan
                    _res[f'{_period}_annual'] = np.nan
                    _res[f'{_period}_mdd'] = np.nan
        
                    
            recent_1w_ret = _res['w1']
            recent_1w_annual_ret = _res['w1_annual']
            recent_1w_mdd = _res['w1_mdd']

            recent_1m_ret = _res['m1']
            recent_1m_annual_ret = _res['m1_annual']
            recent_1m_mdd = _res['m1_mdd']
            
            recent_3m_ret = _res['m3']
            recent_3m_annual_ret = _res['m3_annual']
            recent_3m_mdd = _res['m3_mdd']

            recent_6m_ret = _res['m6']
            recent_6m_annual_ret = _res['m6_annual']
            recent_6m_mdd = _res['m6_mdd']
            
            recent_1y_ret = _res['y1']
            recent_1y_annual_ret = _res['y1_annual']
            recent_1y_mdd = _res['y1_mdd']

            recent_2y_ret = _res['y2']
            recent_2y_annual_ret = _res['y2_annual']
            recent_2y_mdd = _res['y2_mdd']

            recent_3y_ret = _res['y3']
            recent_3y_annual_ret = _res['y3_annual']
            recent_3y_mdd = _res['y3_mdd']

            recent_4y_ret = _res['y4']
            recent_4y_annual_ret = _res['y4_annual']
            recent_4y_mdd = _res['y4_mdd']

            recent_5y_ret = _res['y5']
            recent_5y_annual_ret = _res['y5_annual']
            recent_5y_mdd = _res['y5_mdd']

            if 'w1_begin_date' in _res:
                w1_begin_date = _res['w1_begin_date']
                w1_end_date = _res['w1_end_date']
            else:
                w1_begin_date = sr.index[0]
                w1_end_date = sr.index[-1]

            history_ret = sr.loc[:last_date][-1] / sr[0] - 1
            days = (dates[-1] - dates[0]).days
            history_annual_ret = math.pow(history_ret + 1, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1 if days != 0 else 0
            sr_week = CalculatorBase.data_resample_weekly_nav(sr,rule='W-FRI')
            if sr_week.shape[0] < 5:
                recent_4w_ret = sr_week[-1] / sr_week[0] - 1
            else:
                recent_4w_ret = sr_week[-1] / sr_week[-5] - 1

            _date_20_b = datetime.date(2019,12,31)
            _date_20_e = datetime.date(2020,12,31)
            date_20_b = max(sr.index.min(),_date_20_b)
            date_20_e = min(sr.index.max(),_date_20_e)
            _sr = sr[date_20_b:date_20_e]
            y_2020_ret = _sr[-1] / _sr[0] - 1 if not _sr.empty else None
            lasty_year = sr.index[-1].year - 1
            _date_last_b = datetime.date(lasty_year-1,12,31)
            _date_last_e = datetime.date(lasty_year,12,31)
            date_last_b = max(sr.index.min(),_date_last_b)
            date_last_e = min(sr.index.max(),_date_last_e)
            _sr = sr[date_last_b:date_last_e]
            if not _sr.empty:
                y_last_ret = _sr[-1] / _sr[0] - 1 
                y_last_ret_annual = math.pow(_sr[-1] / _sr[0], CalculatorBase.TOTAL_DAYS_PER_YEAR/(date_last_e-date_last_b).days) - 1
                y_last_mdd = 1 - (_sr / _sr.cummax()).min()
            else:
                y_last_ret = None
                y_last_ret_annual = None
                y_last_mdd = None

            return {
                'recent_natural_year_ret':recent_natural_year_ret,
                'recent_1w_ret':recent_1w_ret,
                'recent_4w_ret':recent_4w_ret,
                'recent_1m_ret':recent_1m_ret,
                'recent_3m_ret':recent_3m_ret,
                'recent_6m_ret':recent_6m_ret,
                'recent_1y_ret':recent_1y_ret,
                'recent_2y_ret':recent_2y_ret,
                'recent_3y_ret':recent_3y_ret,
                'recent_4y_ret':recent_4y_ret,
                'recent_5y_ret':recent_5y_ret,
                'history_ret':history_ret,
                'history_annual_ret':history_annual_ret,
                'y_2020_ret':y_2020_ret,
                'last_year_ret':y_last_ret,
                'w1_begin_date':w1_begin_date,
                'w1_end_date':w1_end_date,
                'recent_1w_annual_ret':recent_1w_annual_ret,
                'recent_1w_mdd':recent_1w_mdd,
                'recent_1m_annual_ret':recent_1m_annual_ret,
                'recent_1m_mdd':recent_1m_mdd,
                'recent_3m_annual_ret':recent_3m_annual_ret,
                'recent_3m_mdd':recent_3m_mdd,
                'recent_6m_annual_ret':recent_6m_annual_ret,
                'recent_6m_mdd':recent_6m_mdd,
                'recent_1y_annual_ret':recent_1y_annual_ret,
                'recent_1y_mdd':recent_1y_mdd,
                'recent_2y_annual_ret':recent_2y_annual_ret,
                'recent_2y_mdd':recent_2y_mdd,
                'recent_3y_annual_ret':recent_3y_annual_ret,
                'recent_3y_mdd':recent_3y_mdd,
                'recent_4y_annual_ret':recent_4y_annual_ret,
                'recent_4y_mdd':recent_4y_mdd,
                'recent_5y_annual_ret':recent_5y_annual_ret,
                'recent_5y_mdd':recent_5y_mdd,
                'y_last_ret_annual':y_last_ret_annual,
                'y_last_mdd':y_last_mdd,
            }
        
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.calc_recent_ret')
            return {
                'recent_natural_year_ret':0,
                'recent_1w_ret':0,
                'recent_4w_ret':0,
                'recent_1m_ret':0,
                'recent_3m_ret':0,
                'recent_6m_ret':0,
                'recent_1y_ret':0,
                'recent_2y_ret':0,
                'recent_3y_ret':0,
                'recent_4y_ret':0,
                'recent_5y_ret':0,
                'history_ret':0,
                'history_annual_ret':0,
                'y_2020_ret':0,
                'last_year_ret':0,
                'w1_begin_date':None,
                'w1_end_date':None,
                'recent_1w_annual_ret':0,
                'recent_1w_mdd':0,
                'recent_1m_annual_ret':0,
                'recent_1m_mdd':0,
                'recent_3m_annual_ret':0,
                'recent_3m_mdd':0,
                'recent_6m_annual_ret':0,
                'recent_6m_mdd':0,
                'recent_1y_annual_ret':0,
                'recent_1y_mdd':0,
                'recent_2y_annual_ret':0,
                'recent_2y_mdd':0,
                'recent_3y_annual_ret':0,
                'recent_3y_mdd':0,
                'recent_4y_annual_ret':0,
                'recent_4y_mdd':0,
                'recent_5y_annual_ret':0,
                'recent_5y_mdd':0,
                'y_last_ret_annual':0,
                'y_last_mdd':0,
            }

    @staticmethod
    def get_period_ret(dates: pd.core.series.Series,
                       values: pd.core.series.Series,
                       frequency: str='3M'):
        assert frequency in ['1D','1W','2W','1M','3M','1Y'], 'frequency provided is not included, must in 1D 1W 2W 1M 3M 1Y'
        if frequency == '1Y':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().bfill()
            sr = sr.reindex([sr.index[0]] + _date).pct_change(1).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['ret']
            return df.sort_index()
        if frequency == '3M':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[3,31],[6,30],[9,30],[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().bfill()
            sr = sr.reindex(_date).pct_change(1).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['ret']
            return df.sort_index()
        sr = pd.Series(values, index=dates).sort_index().dropna()
        sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency,).last().dropna()
        df = pd.DataFrame(sr.pct_change(1).dropna())
        df.columns = ['ret']
        td = df.index
        df.index = [i.date() for i in td]
        return df.sort_index()

    @staticmethod
    def calc_recent_vol(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        frequency:         str='1D'):
        assert frequency in ['1D','1W'], 'frequency provided is not included, must in 1D 1W 2W 1M'
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        _end_dt = dates[-1]
        last_date, begin_date_dic = CalculatorBase.get_begin_date_dic(week_end_date = _end_dt)
        assert len(dates) == len(values), 'date and port_values has different numbers'
        sr = pd.Series(values, index=dates).sort_index().dropna()
        sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
        sr.index = [i.date() for i in sr.index]
        sr_ret = sr.pct_change(1).dropna()
        if frequency == '1D':    
            period_1y = CalculatorBase.TRADING_DAYS_PER_YEAR
        elif frequency == '1W':
            period_1y = CalculatorBase.TOTAL_WEEKS_PER_YEAR
        _begin_date = sr_ret.index[0]
        _natural_year_begin_date = datetime.date(last_date.year,1,1)
        recent_natural_year_vol = sr_ret.loc[_natural_year_begin_date:].std(ddof=1) * np.sqrt(period_1y)
        recent_1w_vol = sr_ret.loc[max(begin_date_dic['w1'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_1m_vol = sr_ret.loc[max(begin_date_dic['m1'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_3m_vol = sr_ret.loc[max(begin_date_dic['m3'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_6m_vol = sr_ret.loc[max(begin_date_dic['m6'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_1y_vol = sr_ret.loc[max(begin_date_dic['y1'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_3y_vol = sr_ret.loc[max(begin_date_dic['y3'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        recent_5y_vol = sr_ret.loc[max(begin_date_dic['y5'],_begin_date):].std(ddof=1) * np.sqrt(period_1y)
        history_vol = sr_ret.std(ddof=1) * np.sqrt(period_1y)
        return {
            'recent_natural_year_vol':recent_natural_year_vol,
            'recent_1w_vol':recent_1w_vol,
            'recent_1m_vol':recent_1m_vol,
            'recent_3m_vol':recent_3m_vol,
            'recent_6m_vol':recent_6m_vol,
            'recent_1y_vol':recent_1y_vol,
            'recent_3y_vol':recent_3y_vol,
            'recent_5y_vol':recent_5y_vol,
            'history_vol':history_vol,
        }

    @staticmethod
    def get_period_vol(dates: pd.core.series.Series,
                       values: pd.core.series.Series,
                       frequency: str='3M'):
        assert frequency in ['1D','1W','2W','1M','3M','1Y'], 'frequency provided is not included, must in 1D 1W 2W 1M 3M 1Y'
        sr = pd.Series(values, index=dates).sort_index().pct_change(1).dropna()
        if frequency == '1D':
            roll_d = CalculatorBase.TRADING_DAYS_PER_YEAR
        elif frequency == '1W':
            roll_d = 5
        elif frequency == '2W':
            roll_d = 10
        elif frequency == '1M':
            roll_d = 20
        elif frequency == '3M':
            roll_d = 60
        elif frequency == '1Y':
            roll_d = CalculatorBase.TRADING_DAYS_PER_YEAR
        if frequency == '1Y':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().pct_change(1)
            sr = sr.rolling(window=roll_d).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
            sr = sr.reindex(_date).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['vol']
            return df.sort_index()
        if frequency == '3M':
            sr = pd.Series(values, index=dates).sort_index().dropna()
            year_list = set(sorted([i.year for i in sr.index]))
            _date = []
            for y in year_list:
                for m_d in [[3,31],[6,30],[9,30],[12,31]]:
                    m = m_d[0]
                    d = m_d[1]
                    dt = datetime.date(y,m,d)
                    _date.append(dt)
            sr = sr.reindex(sr.index.union(_date)).sort_index().ffill().pct_change(1)
            sr = sr.rolling(window=roll_d).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
            sr = sr.reindex(_date).dropna()
            sr = sr[sr!=0]
            df = pd.DataFrame(sr)
            df.columns = ['vol']
            df = df.sort_index()
            today = datetime.datetime.now().date()
            this_year = today.year
            year_list = np.array([datetime.date(this_year, m_d[0], m_d[1]) for m_d in [[3,31],[6,30],[9,30],[12,31]]])
            next_report_date = year_list[year_list>today][0]
            return df.loc[:next_report_date]
        sr = sr.rolling(window=roll_d).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
        sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last().dropna()
        df = pd.DataFrame(sr)
        df.columns = ['vol']
        td = df.index
        df.index = [i.date() for i in td]
        return df.sort_index()

    @staticmethod
    def get_fund_correlation_df(df: pd.DataFrame):
        # 列是资产代码 index是日期
        # 求最近一年日度收益相关性
        return df.pct_change(1).corr()

    @staticmethod
    def get_stat_result(dates:             pd.core.series.Series,
                        values:            pd.core.series.Series,
                        benchmark_values:  pd.core.series.Series=None,
                        risk_free_rate:    float=0.025,
                        frequency:         str='1D',
                        ret_method:        str='log_ret',
                        ):# '1W' '1M'
        assert frequency in ['1D','1W','2W','1M'], 'frequency provided is not included, must in 1D 1W 2W 1M'
        assert ret_method in ['log_ret', 'pct_ret'], 'ret_method is not included'
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        _end_dt = dates[-1]
        last_date, begin_date_dic = CalculatorBase.get_begin_date_dic(week_end_date = _end_dt)
        assert len(dates) == len(values), 'date and port_values has different numbers'
        if len(dates) >= 2:
            sr = pd.Series(values, index=dates).sort_index()
            sr = sr.set_axis(pd.to_datetime(sr.index), inplace=False).resample(frequency).last()
            b_d = sr.dropna().index[0]
            e_d = sr.dropna().index[-1]
            fund_nav_completion = sr.dropna().shape[0] / sr.loc[b_d:e_d].shape[0]
            sr = sr.dropna()
            sr.index = [i.date() for i in sr.index]
            natural_day_lenth = (sr.index[-1] - sr.index[0]).days
            price_num = sr.dropna().shape[0]
            nav_density = price_num / natural_day_lenth if natural_day_lenth !=0 else 0
            nav_type = CalculatorBase.get_nav_type(values=values, dates=dates)
            if frequency == '1D':            
                period_1m = 20
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TRADING_DAYS_PER_YEAR
                risk_free_per_period = risk_free_rate / CalculatorBase.TRADING_DAYS_PER_YEAR
            elif frequency == '1W':
                period_1m = 4
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TOTAL_WEEKS_PER_YEAR
                risk_free_per_period = risk_free_rate / 52
            elif frequency == '2W':
                period_1m = 2
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TOTAL_WEEKS_PER_YEAR / 2
                risk_free_per_period = risk_free_rate / 26
            elif frequency == '1M':
                period_1m = 1
                period_3m = period_1m * 3
                period_6m = period_1m * 6
                period_1y = CalculatorBase.TOTAL_MONTHS_PER_YEAR
                risk_free_per_period = risk_free_rate / 12
            if ret_method == 'log_ret':
                sr_ret = np.log(sr).diff(1).iloc[1:]
            elif ret_method == 'pct_ret':
                sr_ret = sr.pct_change(1).iloc[1:]
            start_date = sr.index[0] 
            end_date = sr.index[-1]
            days = (end_date - start_date).days
            last_unit_nav = sr[-1] / sr[0]
            cumu_ret = last_unit_nav - 1
            trade_month = days / 30
            trade_year = days / CalculatorBase.TOTAL_DAYS_PER_YEAR
            annual_ret = math.pow(last_unit_nav, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1 if days != 0 else 0
            annual_vol = CalculatorBase.calc_modify_annual_vol(sr)
            _begin_date = sr.index[0]
            recent_ret_res = CalculatorBase.calc_recent_ret(dates=dates,values=values,frequency='1D')
            recent_year_ret = recent_ret_res['recent_natural_year_ret']
            recent_1w_ret = recent_ret_res['recent_1w_ret']
            recent_1m_ret = recent_ret_res['recent_1m_ret']
            recent_3m_ret = recent_ret_res['recent_3m_ret']
            recent_6m_ret = recent_ret_res['recent_6m_ret']
            recent_1y_ret = recent_ret_res['recent_1y_ret']
            recent_3y_ret = recent_ret_res['recent_3y_ret']
            recent_5y_ret = recent_ret_res['recent_5y_ret']
            recent_4w_ret = recent_ret_res['recent_4w_ret']
            last_year_ret = recent_ret_res['last_year_ret']
            w1_begin_date = recent_ret_res['w1_begin_date']
            recent_2y_ret = recent_ret_res['recent_2y_ret']
            recent_1w_annual_ret = recent_ret_res['recent_1w_annual_ret']
            recent_1w_mdd = recent_ret_res['recent_1w_mdd']
            recent_1m_annual_ret = recent_ret_res['recent_1m_annual_ret']
            recent_1m_mdd = recent_ret_res['recent_1m_mdd']
            recent_3m_annual_ret = recent_ret_res['recent_3m_annual_ret']
            recent_3m_mdd = recent_ret_res['recent_3m_mdd']
            recent_6m_annual_ret = recent_ret_res['recent_6m_annual_ret']
            recent_6m_mdd = recent_ret_res['recent_6m_mdd']
            recent_1y_annual_ret = recent_ret_res['recent_1y_annual_ret']
            recent_1y_mdd = recent_ret_res['recent_1y_mdd']
            recent_2y_annual_ret = recent_ret_res['recent_2y_annual_ret']
            recent_2y_mdd = recent_ret_res['recent_2y_mdd']
            recent_3y_annual_ret = recent_ret_res['recent_3y_annual_ret']
            recent_3y_mdd = recent_ret_res['recent_3y_mdd']
            recent_4y_annual_ret = recent_ret_res['recent_4y_annual_ret']
            recent_4y_mdd = recent_ret_res['recent_4y_mdd']
            recent_5y_annual_ret = recent_ret_res['recent_5y_annual_ret']
            recent_5y_mdd = recent_ret_res['recent_5y_mdd']
            y_last_ret_annual = recent_ret_res['y_last_ret_annual']
            y_last_mdd = recent_ret_res['y_last_mdd']

            w1_end_date = recent_ret_res['w1_end_date']
            y_2020_ret = recent_ret_res['y_2020_ret']
            worst_3m_ret = sr.pct_change(period_3m).min()
            worst_6m_ret = sr.pct_change(period_6m).min()
            last_mv_diff = sr[-1] - sr[-2]
            last_increase_rate = (sr[-1] - sr[-2])/ sr[-2]
            sharpe = (annual_ret - risk_free_rate) / annual_vol
            recent_drawdown = 1 - (sr[-1] / sr.max())
            if np.isnan(recent_drawdown):
                recent_mdd_date1 = None
                recent_mdd_lens = None
            else:
                recent_mdd_date1 = sr.idxmax()
                recent_mdd_lens = (end_date - recent_mdd_date1).days
            mdd_part =  sr / sr.rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            if np.isnan(mdd):
                mdd_date2 = None
                mdd_date1 = None
                mdd_lens = None
            else:
                mdd_date2 = mdd_part.idxmin()
                mdd_date1 = sr[:mdd_date2].idxmax()
                mdd_lens = (mdd_date2-mdd_date1).days
            calmar = (annual_ret - risk_free_rate) / mdd if mdd != 0 else 0
            sr_ret_no_risk_free = sr_ret - risk_free_rate / period_1y
            downside_risk = np.minimum(sr_ret_no_risk_free,0).std(ddof=1) * np.sqrt(period_1y)
            sortino = (annual_ret - risk_free_rate) / downside_risk
            var =  np.quantile(sr_ret, 0.05) # 按照最大回撤处理， 损失非负， 计算ervar用
            cvar = max(0, -sr_ret[sr_ret < var].mean())
            var = - var
            ervar = (annual_ret - risk_free_rate) / var
            skew_value = skew(sr_ret)
            kurtosis_value = kurtosis(sr_ret)        
            sr_ret_monthly = CalculatorBase.data_resample_monthly_ret(df=sr_ret, min_count=1).dropna()
            raise_month_num = sr_ret_monthly[sr_ret_monthly>0].shape[0]
            drop_month_num = sr_ret_monthly[sr_ret_monthly<0].shape[0]
            mdd_recover, mdd_recover_date1, mdd_recover_date2, mdd_recover_lens = CalculatorBase.mdd_recover_analysis(values=values, dates=dates)
            win_rate_0_period = np.mean((sr_ret > 0) * 1)

            if benchmark_values is not None:
                assert len(dates) == len(benchmark_values), 'date and bench_values has different numbers'
                sr_benchmark = pd.Series(benchmark_values, index=dates).sort_index().dropna()
                sr_benchmark = sr_benchmark.set_axis(pd.to_datetime(sr_benchmark.index), inplace=False).resample(frequency).last().dropna()
                total_relative_ret = sr.iloc[-1]/ sr.iloc[0] - sr_benchmark.iloc[-1] / sr_benchmark.iloc[0]
                sr_benchmark.index = [i.date() for i in sr_benchmark.index]
                if ret_method == 'log_ret':
                    sr_benchmark_ret = np.log(sr_benchmark).diff(1).iloc[1:]
                elif ret_method == 'pct_ret':
                    sr_benchmark_ret = sr_benchmark.pct_change(1).iloc[1:]
                last_unit_benchmark = sr_benchmark[-1] / sr_benchmark[0]
                annual_ret_benchmark = math.pow(last_unit_benchmark, CalculatorBase.TOTAL_DAYS_PER_YEAR / days) - 1
                df = pd.concat([sr, sr_benchmark],axis=1)
                df.columns= ['fund_ret','index_ret']
                df_monthly = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule='1M').last()
                df_monthly.index = df_monthly.index.date
                monthly_ret = df_monthly.pct_change(1).dropna()
                continue_value_m = CalculatorBase.calc_continue_regress_v(monthly_ret.fund_ret,risk_free_rate)
                df_ret_risk_free = (df.pct_change(1) - risk_free_per_period).dropna()
                cl_res = CalculatorBase.calc_cl_alpha_beta(df_ret_risk_free.to_numpy())
                alpha_cl = cl_res['alpha'] * period_1y
                beta_cl = cl_res['beta']
                beta_bear = cl_res['beta_bear']
                beta_bull = cl_res['beta_bull']
                hm_res = CalculatorBase.calc_hm_alpha_beta(df_ret_risk_free.to_numpy())
                alpha_hm = hm_res['alpha'] * period_1y
                beta_hm = hm_res['beta']
                excess_ret = annual_ret - annual_ret_benchmark
                _index = sorted(sr_ret.index.intersection(sr_benchmark_ret.index))
                sr_ret = sr_ret.reindex(_index)
                sr_benchmark_ret = sr_benchmark_ret.reindex(_index)
                sr_ret_weekly = CalculatorBase.data_resample_weekly_ret(df=sr_ret,min_count=1).dropna()
                sr_benchmark_ret_weekly = CalculatorBase.data_resample_weekly_ret(df=sr_benchmark_ret, min_count=1).dropna()
                sr_benchmark_ret_monthly = CalculatorBase.data_resample_monthly_ret(df=sr_benchmark_ret, min_count=1).dropna()
                
                _sr_benchmark = sr_benchmark.loc[mdd_date1:mdd_date2]
                mdd_benchmark_same_time = 1 - (_sr_benchmark / _sr_benchmark.cummax()).min()
                asset_to_bmk_mdd = mdd / mdd_benchmark_same_time
                win_rate = np.mean((sr_ret > sr_benchmark_ret) * 1)
                win_rate_0 = np.mean((sr_ret > 0) * 1)
                win_rate_weekly = np.mean((sr_ret_weekly > sr_benchmark_ret_weekly) * 1)
                win_rate_0_weekly = np.mean((sr_ret_weekly > 0) * 1)
                win_rate_monthly = np.mean((sr_ret_monthly > sr_benchmark_ret_monthly) * 1)
                win_rate_0_monthly = np.mean((sr_ret_monthly > 0) * 1)
                
                corr = sr_ret.corr(sr_benchmark_ret)    
                track_err = (sr_ret - sr_benchmark_ret).std(ddof=1) * np.sqrt(period_1y)
                info = excess_ret / track_err
                x = (sr_benchmark_ret - risk_free_per_period).dropna() .values
                y = (sr_ret - risk_free_per_period).dropna().values
                x = sm.add_constant(x)
                model = regression.linear_model.OLS(y,x).fit()
                alpha = model.params[0]
                beta = model.params[1]
                treynor = (annual_ret - risk_free_rate) / beta
                not_system_risk = sum((sr_ret - risk_free_per_period)**2) \
                                - alpha * sum(sr_ret - risk_free_per_period) \
                                - beta * sum((sr_ret - risk_free_per_period) * (sr_benchmark_ret - risk_free_per_period))
                not_system_risk = np.sqrt(not_system_risk / (len(sr_ret) - 2))
                alpha = alpha * period_1y
                df_ret = np.log(df).diff(1).iloc[1:]
                df_ret_up = df_ret[df_ret.index_ret > 0] + 1 
                up_cap_top = df_ret_up.fund_ret.product() ** (period_1y/len(df_ret_up)) - 1
                up_cap_bot = df_ret_up.index_ret.product() ** (period_1y/len(df_ret_up)) - 1
                up_capture = up_cap_top / up_cap_bot

                df_ret_down = df_ret[df_ret.index_ret < 0] + 1
                df_ret_down_length = max(1,len(df_ret_down))
                down_cap_top = df_ret_down.fund_ret.product() ** (period_1y/df_ret_down_length) - 1
                down_cap_bot = df_ret_down.index_ret.product() ** (period_1y/df_ret_down_length) - 1
                down_capture = down_cap_top / down_cap_bot

                df_ret['excess_ret'] = df_ret.fund_ret - df_ret.index_ret
                df_ret = df_ret.replace(np.Inf,np.nan).replace(-np.Inf,np.nan).dropna()
                if df_ret.shape[0] < 4:
                    excess_ret_cotinue = np.nan
                else:
                    mod = AutoReg(endog=df_ret.excess_ret.values, lags=1)
                    res = mod.fit()
                    excess_ret_cotinue = res.params[0]

                sr_excess_value = np.exp((sr_ret-sr_benchmark_ret).cumsum())
                _b_d = sr.loc[:sr_excess_value.index[0]].index[-2]
                sr_excess_value.loc[_b_d] = 1
                sr_excess_value = sr_excess_value.sort_index()
                mdd_part =  sr_excess_value / sr_excess_value.rolling(window=sr_excess_value.shape[0], min_periods=1).max()
                excess_mdd = 1 - mdd_part.min()
                
                
                if np.isnan(excess_mdd):
                    excess_mdd_date2 = None
                    excess_mdd_date1 = None
                    excess_mdd_lens = None
                else:
                    excess_mdd_date2 = mdd_part.idxmin()
                    excess_mdd_date1 = sr[:excess_mdd_date2].idxmax()
                    excess_mdd_lens = (excess_mdd_date2-excess_mdd_date1).days
                
                return {
                    'start_date':start_date,
                    'end_date':end_date,
                    'trade_month':trade_month,
                    'trade_year':trade_year,
                    'last_unit_nav':last_unit_nav,
                    'cumu_ret':cumu_ret,
                    'annual_ret':annual_ret,
                    'annual_vol':annual_vol,
                    'sharpe':sharpe,
                    'fund_nav_completion':fund_nav_completion,
                    'recent_year_ret':recent_year_ret,
                    'recent_1w_ret':recent_1w_ret,
                    'recent_1m_ret':recent_1m_ret,
                    'recent_4w_ret':recent_4w_ret,
                    'y_2020_ret':y_2020_ret,
                    'last_year_ret':last_year_ret,
                    'recent_3m_ret':recent_3m_ret,
                    'recent_6m_ret':recent_6m_ret,
                    'recent_1y_ret':recent_1y_ret,
                    'recent_3y_ret': recent_3y_ret,
                    'recent_5y_ret':recent_5y_ret,
                    'worst_3m_ret':worst_3m_ret,
                    'worst_6m_ret':worst_6m_ret,
                    'last_mv_diff':last_mv_diff,
                    'last_increase_rate':last_increase_rate,
                    'recent_drawdown':recent_drawdown,
                    'recent_mdd_date1':recent_mdd_date1,
                    'recent_mdd_lens':recent_mdd_lens,
                    'mdd':mdd,
                    'mdd_date1':mdd_date1,
                    'mdd_date2':mdd_date2,
                    'mdd_lens':mdd_lens,
                    'calmar':calmar,
                    'downside_risk':downside_risk,
                    'sortino':sortino,
                    'var':var,
                    'ervar':ervar,
                    'skew':skew_value,
                    'kurtosis':kurtosis_value,
                    'alpha_cl':alpha_cl,
                    'beta_cl':beta_cl,
                    'beta_bear':beta_bear,
                    'beta_bull':beta_bull,
                    'alpha_hm':alpha_hm,
                    'beta_hm':beta_hm,
                    'excess_ret':excess_ret,
                    'win_rate':win_rate,
                    'win_rate_0':win_rate_0,
                    'win_rate_weekly':win_rate_weekly,
                    'win_rate_0_weekly':win_rate_0_weekly,
                    'win_rate_monthly':win_rate_monthly,
                    'win_rate_0_monthly':win_rate_0_monthly,
                    'raise_month_num':raise_month_num,
                    'drop_month_num':drop_month_num,
                    'corr':corr,
                    'track_err':track_err,
                    'info':info,
                    'alpha':alpha,
                    'beta':beta,
                    'treynor':treynor,
                    'not_system_risk':not_system_risk,
                    'cvar':cvar,
                    'up_capture':up_capture,
                    'down_capture':down_capture,
                    'excess_ret_cotinue':excess_ret_cotinue,
                    'continue_value_m':continue_value_m,
                    'excess_mdd':excess_mdd,
                    'excess_mdd_date1':excess_mdd_date1,
                    'excess_mdd_date2':excess_mdd_date2,
                    'excess_mdd_lens':excess_mdd_lens,
                    'total_relative_ret':total_relative_ret,
                    'mdd_recover':mdd_recover,
                    'mdd_recover_date1':mdd_recover_date1,
                    'mdd_recover_date2':mdd_recover_date2,
                    'mdd_recover_lens':mdd_recover_lens,
                    'win_rate_0_period':win_rate_0_period,
                    'nav_density':nav_density,
                    'nav_type':nav_type,
                    'w1_begin_date':w1_begin_date,
                    'w1_end_date':w1_end_date,
                    'asset_to_bmk_mdd':asset_to_bmk_mdd,
                    'recent_2y_ret':recent_2y_ret,
                    'recent_1w_annual_ret':recent_1w_annual_ret,
                    'recent_1w_mdd':recent_1w_mdd,
                    'recent_1m_annual_ret':recent_1m_annual_ret,
                    'recent_1m_mdd':recent_1m_mdd,
                    'recent_3m_annual_ret':recent_3m_annual_ret,
                    'recent_3m_mdd':recent_3m_mdd,
                    'recent_6m_annual_ret':recent_6m_annual_ret,
                    'recent_6m_mdd':recent_6m_mdd,
                    'recent_1y_annual_ret':recent_1y_annual_ret,
                    'recent_1y_mdd':recent_1y_mdd,
                    'recent_2y_annual_ret':recent_2y_annual_ret,
                    'recent_2y_mdd':recent_2y_mdd,
                    'recent_3y_annual_ret':recent_3y_annual_ret,
                    'recent_3y_mdd':recent_3y_mdd,
                    'recent_4y_annual_ret':recent_4y_annual_ret,
                    'recent_4y_mdd':recent_4y_mdd,
                    'recent_5y_annual_ret':recent_5y_annual_ret,
                    'recent_5y_mdd':recent_5y_mdd,
                    'y_last_ret_annual':y_last_ret_annual,
                    'y_last_mdd':y_last_mdd,
                }
            else:
                return {
                    'start_date':start_date,
                    'end_date':end_date,
                    'trade_month':trade_month,
                    'trade_year':trade_year,
                    'last_unit_nav':last_unit_nav,
                    'cumu_ret':cumu_ret,
                    'annual_ret':annual_ret,
                    'annual_vol':annual_vol,
                    'sharpe':sharpe,
                    'fund_nav_completion':fund_nav_completion,
                    'recent_year_ret':recent_year_ret,
                    'recent_1w_ret':recent_1w_ret,
                    'recent_1m_ret':recent_1m_ret,
                    'recent_4w_ret':recent_4w_ret,
                    'y_2020_ret':y_2020_ret,
                    'last_year_ret':last_year_ret,
                    'recent_3m_ret':recent_3m_ret,
                    'recent_6m_ret':recent_6m_ret,
                    'recent_1y_ret':recent_1y_ret,
                    'recent_3y_ret': recent_3y_ret,
                    'recent_5y_ret':recent_5y_ret,
                    'worst_3m_ret':worst_3m_ret,
                    'worst_6m_ret':worst_6m_ret,
                    'last_mv_diff':last_mv_diff,
                    'last_increase_rate':last_increase_rate,
                    'recent_drawdown':recent_drawdown,
                    'recent_mdd_date1':recent_mdd_date1,
                    'recent_mdd_lens':recent_mdd_lens,
                    'raise_month_num':raise_month_num,
                    'drop_month_num':drop_month_num,
                    'mdd':mdd,
                    'mdd_date1':mdd_date1,
                    'mdd_date2':mdd_date2,
                    'mdd_lens':mdd_lens,
                    'calmar':calmar,
                    'downside_risk':downside_risk,
                    'sortino':sortino,
                    'var':var,
                    'cvar':cvar,
                    'ervar':ervar,
                    'skew':skew_value,
                    'kurtosis':kurtosis_value,
                    'mdd_recover':mdd_recover,
                    'mdd_recover_date1':mdd_recover_date1,
                    'mdd_recover_date2':mdd_recover_date2,
                    'mdd_recover_lens':mdd_recover_lens,
                    'win_rate_0_period':win_rate_0_period,
                    'nav_density':nav_density,
                    'nav_type':nav_type,
                    'w1_begin_date':w1_begin_date,
                    'w1_end_date':w1_end_date,
                    'recent_2y_ret':recent_2y_ret,
                    'recent_1w_annual_ret':recent_1w_annual_ret,
                    'recent_1w_mdd':recent_1w_mdd,
                    'recent_1m_annual_ret':recent_1m_annual_ret,
                    'recent_1m_mdd':recent_1m_mdd,
                    'recent_3m_annual_ret':recent_3m_annual_ret,
                    'recent_3m_mdd':recent_3m_mdd,
                    'recent_6m_annual_ret':recent_6m_annual_ret,
                    'recent_6m_mdd':recent_6m_mdd,
                    'recent_1y_annual_ret':recent_1y_annual_ret,
                    'recent_1y_mdd':recent_1y_mdd,
                    'recent_2y_annual_ret':recent_2y_annual_ret,
                    'recent_2y_mdd':recent_2y_mdd,
                    'recent_3y_annual_ret':recent_3y_annual_ret,
                    'recent_3y_mdd':recent_3y_mdd,
                    'recent_4y_annual_ret':recent_4y_annual_ret,
                    'recent_4y_mdd':recent_4y_mdd,
                    'recent_5y_annual_ret':recent_5y_annual_ret,
                    'recent_5y_mdd':recent_5y_mdd,
                    'y_last_ret_annual':y_last_ret_annual,
                    'y_last_mdd':y_last_mdd,
                }
        else:
            return {
                    'start_date':dates[0] if len(dates) > 0 else None,
                    'end_date':dates[-1] if len(dates) > 0 else None,
                    'trade_month':0,
                    'trade_year':0,
                    'last_unit_nav':None,
                    'cumu_ret':0,
                    'annual_ret':0,
                    'annual_vol':0,
                    'fund_nav_completion':0,
                    'sharpe':0,
                    'recent_year_ret':0,
                    'recent_1w_ret':0,
                    'recent_1m_ret':0,
                    'recent_4w_ret':0,
                    'y_2020_ret':0,
                    'last_year_ret':0,
                    'recent_3m_ret':0,
                    'recent_6m_ret':0,
                    'recent_1y_ret':0,
                    'recent_3y_ret': 0,
                    'recent_5y_ret':0,
                    'worst_3m_ret':0,
                    'worst_6m_ret':0,
                    'last_mv_diff':0,
                    'last_increase_rate':0,
                    'recent_drawdown':0,
                    'recent_mdd_date1':dates[0] if len(dates) > 0 else None,
                    'recent_mdd_lens':0,
                    'mdd':0,
                    'mdd_date1':dates[0] if len(dates) > 0 else None,
                    'mdd_date2':dates[0] if len(dates) > 0 else None,
                    'mdd_lens':0,
                    'calmar':0,
                    'downside_risk':0,
                    'sortino':0,
                    'var':0,
                    'cvar':0,
                    'ervar':0,
                    'skew':0,
                    'kurtosis':0,
                    'alpha_cl':0,
                    'beta_cl':0,
                    'beta_bear':0,
                    'beta_bull':0,
                    'alpha_hm':0,
                    'beta_hm':0,
                    'excess_ret':0,
                    'win_rate':0,
                    'win_rate_0':0,
                    'win_rate_weekly':0,
                    'win_rate_0_weekly':0,
                    'win_rate_monthly':0,
                    'win_rate_0_monthly':0,
                    'raise_month_num':0,
                    'drop_month_num':0,
                    'raise_month_num':0,
                    'drop_month_num':0,
                    'corr':0,
                    'track_err':0,
                    'info':0,
                    'alpha':0,
                    'beta':0,
                    'treynor':0,
                    'not_system_risk':0,
                    'cvar':0,
                    'up_capture':0,
                    'down_capture':0,
                    'excess_ret_cotinue':0,
                    'continue_value_m':0,
                    'excess_mdd':0,
                    'excess_mdd_date1':dates[0] if len(dates) > 0 else None,
                    'excess_mdd_date2':dates[0] if len(dates) > 0 else None,
                    'excess_mdd_lens':0,
                    'total_relative_ret':0,
                    'mdd_recover':0,
                    'mdd_recover_date1':dates[0] if len(dates) > 0 else None,
                    'mdd_recover_date2':dates[0] if len(dates) > 0 else None,
                    'mdd_recover_lens':0,
                    'win_rate_0_period':0,
                    'nav_density':0,
                    'nav_type':'nan',
                    'w1_begin_date':dates[0] if len(dates) > 0 else None,
                    'w1_end_date':dates[0] if len(dates) > 0 else None,
                    'recent_2y_ret':0,
                    'recent_1w_annual_ret':0,
                    'recent_1w_mdd':0,
                    'recent_1m_annual_ret':0,
                    'recent_1m_mdd':0,
                    'recent_3m_annual_ret':0,
                    'recent_3m_mdd':0,
                    'recent_6m_annual_ret':0,
                    'recent_6m_mdd':0,
                    'recent_1y_annual_ret':0,
                    'recent_1y_mdd':0,
                    'recent_2y_annual_ret':0,
                    'recent_2y_mdd':0,
                    'recent_3y_annual_ret':0,
                    'recent_3y_mdd':0,
                    'recent_4y_annual_ret':0,
                    'recent_4y_mdd':0,
                    'recent_5y_annual_ret':0,
                    'recent_5y_mdd':0,
                    'y_last_ret_annual':0,
                    'y_last_mdd':0,
                    'asset_to_bmk_mdd':0,
                }


    @staticmethod    
    def mdd_recover_analysis(values,dates):
        sr = pd.Series(values, index=dates).sort_index()
        if sr.empty:
            mdd = 0
            mdd_date1 = None
            mdd_date2 = None
            mdd_lens = 0
            return mdd, mdd_date1, mdd_date2, mdd_lens
        mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
        mdd = 1 - mdd_part.min()
        if mdd == 0:
            mdd_date1 = None
            mdd_date2 = None
            mdd_lens = 0
        else:
            mdd_date = mdd_part.idxmin()
            mdd_date1 = sr[:mdd_date].idxmax()
            sr_tmp = sr[mdd_date1:]
            recover_sr = sr_tmp[sr_tmp> sr[mdd_date1]]
            if recover_sr.empty:
                mdd_date2 = sr_tmp.index[-1]
            else: 
                mdd_date2 = sr_tmp[sr_tmp> sr[mdd_date1]].index[0]
            mdd_lens = sr.loc[mdd_date1:mdd_date2].shape[0]
        return mdd, mdd_date1, mdd_date2, mdd_lens

    @staticmethod
    def simulated_mv_calc(df,fee_dic,weight_dic):
        '''
        df structure: 
                    index   : datetime
                    column  : different assets mv
        i.g.

                	A	    B	    C	    D	    E
        2020-03-02	1.2047	1.0858	1.4002	1.0482	1.000
        2020-03-06	1.2156	1.0983	1.4211	1.0496	1.005
        
        mng_fee_dic: 
        i.g.            
            fee_dic = {'A':0.2,'B':0.2,'C':0.2,'D':0.2,'E':0.2}
        
        weight_dic: value or weight 
        i.g.
            weight_dic = {'A':1,'B':1,'C':1,'D':1,'E':1}

        '''
        # 归一化
        df_to_1 = df / df.iloc[0]
        # 盈利
        excess_ret = (df_to_1 - 1)
        # 盈 或 亏
        earn_con = (excess_ret > 0) * 1
        # 费
        pay_mng_fee = excess_ret * earn_con * pd.DataFrame([fee_dic]).values
        # 净值
        simu_nav = df_to_1
        simu_nav_vir = df_to_1 - pay_mng_fee
        # 真实总净值
        simu_nav_total = pd.DataFrame((simu_nav*pd.DataFrame([weight_dic]).values).sum(axis=1))
        # 虚拟总净值
        simu_nav_vir_total = pd.DataFrame((simu_nav_vir*pd.DataFrame([weight_dic]).values).sum(axis=1))
        simu_nav_total.columns = ['mv']
        simu_nav_vir_total.columns = ['mv']
        return simu_nav_vir_total, simu_nav_total

    @staticmethod
    def df_data_select(df, start_date, end_date, cols, resample_method, drop_col_title):
        # 组合净值分析时 净值选取
        # 设置 起始日 截止日 选取列名 对列共同名字是否去掉
        if end_date is None:
            end_date = datetime.date(2040,1,1)
        if start_date is None:
            start_date = datetime.date(2010,1,1)
        if cols is not None:
            df = df[cols]
        df = CalculatorBase.data_resample(df.loc[start_date:end_date], resample_method, drop_col_title)
        return df

    @staticmethod
    def data_resample(df, resample_method, drop_col_title=True):
        COL_TITLE = '融智'
        if resample_method == '1D':
            return df 
        if drop_col_title:
            cols = df.columns.tolist()
            cols = [i.replace(COL_TITLE,'') for i in cols]
            df.columns = cols
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(resample_method).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df.ffill()

    @staticmethod
    def proper_round(num, dec=0):
        dec = Decimal(10) ** -dec
        return float(Decimal(str(num)).quantize(dec, context=Context(rounding=ROUND_HALF_UP)))

    @staticmethod
    def stock_ipo_continue_up_limit(    dates:  pd.core.series.Series,
                                        values: pd.core.series.Series,
                                        ipo_price: float,
                                        list_date: datetime.date,
                                        ipo_date: datetime.date,
                                        trade_market: str,
                                        stock_name: str,
                                        stock_code: str):
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        assert len(dates) == len(values), 'date and port_values has different numbers'
        df = pd.DataFrame(values, index=dates).sort_index().dropna()
        col = 'price'
        df.columns = [col]
        rate = [1.44] + df.shape[0] * [1.1]
        df.loc[ipo_date,col] = ipo_price
        df = df.sort_index()
        if trade_market in ['上交所主板','深交所中小板']:
            df.loc[:,'limit_price'] = (df.price * rate).map(lambda x: CalculatorBase.proper_round(x,2)).shift(1)
            df.loc[:,'limit_con'] = (df.limit_price == df.price) * 1  
            limit_list = df.limit_con.iloc[1:].tolist()
            if 0 not in limit_list:
                ipo_lens = len(limit_list)
            else:
                ipo_lens = limit_list.index(0)
            num = CalculatorBase.STOCK_IPO_NUM_MARKET[trade_market]
            earning_per_stick = int(round((df.price.values[ipo_lens] - df.price.values[0]) * num))
        else:
            ipo_lens = np.nan
            earning_per_stick = np.nan
        res = {
            'stock_code':stock_code,
            'stock_name':stock_name,
            'trade_market':trade_market,
            'list_date':list_date,
            'ipo_price':ipo_price,
            'latest_price':df.price.values[-1],
            'latest_rate':df.price.pct_change(1).values[-1],
            'cumulate_rate':df.price.values[-1] / df.price.values[0] - 1,
            'first_date_rate':df.price.values[1] / df.price.values[0] - 1,
            'continue_limited_rise':ipo_lens,
            'earn_per_stick':earning_per_stick,
        }
        return res

    @staticmethod
    def conv_bond_ipo_result(   bond_id:str,
                                bond_name:str,
                                list_date:datetime.date,
                                ipo_price:float,
                                dates:pd.core.series.Series,
                                close_values:pd.core.series.Series,
                                open_values:pd.core.series.Series,
                                rate:float,
                                limit_num: Optional[str]):

        rate = rate / 100
        dates = pd.to_datetime(dates)
        dates = [i.date() for i in dates]
        assert len(dates) == len(close_values), 'date and port_values has different numbers'
        assert len(dates) == len(open_values), 'date and port_values has different numbers'
        df = pd.DataFrame({'close':close_values,'open':open_values}, index=dates).sort_index().dropna()
        if limit_num is not None:
            limit_num = int(limit_num.split('网上申购数量上限：')[1].split('手')[0])
        trade_market = bond_id.split('.')[1]
        new_price = df.close[-1]
        total_ret = df.close[-1] / ipo_price - 1
        first_open = df.open[0]
        first_date_rate = df.close[0] / 100 - 1
        earn_per_stick = (first_open - ipo_price) * 10
        if limit_num is not None:
            earn_per_account = earn_per_stick * limit_num * rate
        else:
            earn_per_account = np.nan
        res = {
            'bond_id':bond_id,
            'bond_name':bond_name,
            'list_date':list_date,
            'ipo_price':ipo_price,
            'trade_market':trade_market,
            'new_price':new_price,
            'total_ret':total_ret,
            'earn_per_account':earn_per_account,
            'earn_per_stick':earn_per_stick,
            'first_date_rate':first_date_rate,
            'first_open':first_open,
            'rate':rate,
            'limit_num':limit_num,
        }
        return res
    
    @staticmethod
    def data_resample_monthly_ret(df, rule='1M', min_count=15):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_weekly_ret(df, rule='1W', min_count=3):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_monthly_nav(df, rule='1M'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_weekly_nav(df, rule='W-FRI'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def get_nav_type(values, dates):
        import json
        import requests
        sr = pd.Series(values, index=dates).sort_index()
        tds = pd.to_datetime(sr.index)
        tds = [_.date() for _ in tds]
        sr.index = tds
        b_d = sr.dropna().index[0]
        e_d = sr.dropna().index[-1]
        url = 'http://69.231.155.101:8006/api/v1/basic/trade_dates'
        response = requests.get(url)    
        dts = json.loads(response._content)['data']['datetime']
        dts = pd.to_datetime(dts)
        dts = np.array([i.date() for i in dts])
        dts = dts[(dts>=b_d) & (dts<=e_d)].tolist()
        sr_daily = sr.reindex(dts)
        sr_week = CalculatorBase.data_resample_weekly_nav(df=sr_daily)
        sr_month = CalculatorBase.data_resample_monthly_nav(df=sr_daily)

        nav_density_daily = 1 - (sum(sr_daily.isnull() * 1) / sr_daily.shape[0])
        nav_density_weeky = 1 - (sum(sr_week.isnull() * 1) / sr_week.shape[0])
        nav_density_monthy = 1 - (sum(sr_month.isnull() * 1) / sr_month.shape[0])

        # 如果周度净值价格都在非交易日上，比如周日
        if sr.dropna().shape[0] > 0 and nav_density_monthy == 0:
            sr = pd.Series(values, index=dates).sort_index()
            sr_week = CalculatorBase.data_resample_weekly_nav(df=sr)
            sr_week = sr.reindex(dts)
            sr_week = CalculatorBase.data_resample_weekly_nav(df=sr)
            sr_month = CalculatorBase.data_resample_monthly_nav(df=sr_week)
            nav_density_weeky = 1 - (sum(sr_week.isnull() * 1) / sr_week.shape[0])
            nav_density_monthy = 1 - (sum(sr_month.isnull() * 1) / sr_month.shape[0])
            
        if nav_density_daily > 0.8:
            fund_nav_type = 'daily'
        elif nav_density_weeky > 0.8:
            fund_nav_type = 'weekly'
        else:
            fund_nav_type = 'monthly'
        return fund_nav_type



import dataclasses
import datetime
import pandas
import numpy as np
import math
import statsmodels.api as sm
from statsmodels import regression
from scipy.stats import gmean
from functools import partial
import statsmodels.api as sm
from statsmodels.tsa.ar_model import AutoReg
from ..data.struct import TAAParam, TaaTunerParam, AssetWeight
from .calculator_item import CalculatorBase

@dataclasses.dataclass
class SeriesStatUnit:

    start_date: datetime.date           # 序列的开始时间
    end_date: datetime.date             # 序列的结束时间
    trade_year: float                   # 交易年限
    trade_month:float                   # 交易月度
    last_unit_nav: float                # 末期单位净值
    annualized_ret: float               # 年化收益
    cumu_ret: float                     # 累计收益
    annualized_vol: float               # 年化波动率
    fund_nav_completion:float           # 净值完整度
    sharpe: float                       # 夏普率
    recent_year_ret: float              # 最近自然年收益
    recent_1w_ret: float                # 最近1周收益
    recent_1m_ret: float                # 最近1月收益
    recent_4w_ret: float                # 最近4周收益
    recent_3m_ret: float                # 最近3月收益
    recent_6m_ret: float                # 最近6月收益
    y_2020_ret: float                   # 2020年收益
    last_year_ret: float                # 去年收益
    recent_y1_ret: float                # 近一年收益
    recent_y2_ret: float                # 近二年收益
    recent_y3_ret: float                # 近三年收益
    recent_y5_ret: float                # 近五年收益
    mdd: float                          # 最大回撤
    mdd_date1: datetime.date            # 最大回撤开始时间
    mdd_date2: datetime.date            # 最大回撤结束时间
    mdd_lens: int                       # 最大回撤区间持续时间
    ret_over_mdd: float                 # Calmar 年化收益除以最大回撤
    last_mv_diff: float                 # 末期市值增量              
    last_increase_rate: float           # 末期涨幅
    worst_3m_ret: float                 # 最差三月收益
    worst_6m_ret: float                 # 最差六月收益
    recent_drawdown: float              # 最近回撤
    recent_mdd_date1: datetime.date     # 最近回撤开始时间
    downside_std :float                 # 下行风险
    var : float                         # 风险价值VaR
    cvar: float                         # 条件风险价值 CVaR
    raise_month_num: int                # 上涨月份数
    drop_month_num: int                 # 下跌月份数
    mdd_recover: float                  # 最大回撤恢复
    mdd_recover_date1: datetime.date    # 最大回撤恢复开始时间
    mdd_recover_date2: datetime.date    # 最大回测恢复结束时间
    mdd_recover_lens: int               # 最大回测哈恢复天数
    win_rate_0_period: float            # 设定周期对零胜率
    nav_density: float                  # 净值密度
    nav_type: str                       # 净值类型
    w1_end_date: datetime.date          # 近一周结束时间
    w1_begin_date: datetime.date        # 近一周开始时间
    recent_1w_annual_ret: float         # 近一周年化收益
    recent_1w_mdd: float                # 近一周最大回撤
    recent_1m_annual_ret: float         # 近一月年化收益
    recent_1m_mdd: float                # 近一月最大回撤
    recent_3m_annual_ret: float         # 近三月年化收益
    recent_3m_mdd: float                # 近三月最大回撤
    recent_6m_annual_ret: float         # 近六月年化收益
    recent_6m_mdd: float                # 近六月最大回撤
    recent_1y_annual_ret: float         # 近一年年化收益
    recent_1y_mdd: float                # 近一年最大回撤
    recent_2y_annual_ret: float         # 近两年年化收益
    recent_2y_mdd: float                # 近两年最大回撤
    recent_3y_annual_ret: float         # 近三年年化收益
    recent_3y_mdd: float                # 近三年最大回撤
    recent_4y_annual_ret: float         # 近四年年化收益
    recent_4y_mdd: float                # 近四年最大回撤
    recent_5y_annual_ret: float         # 近五年年化收益
    recent_5y_mdd: float                # 近五年最大回撤
    y_last_ret_annual: float            # 本自然年年化收益
    y_last_mdd: float                   # 本自然年最大回撤


@dataclasses.dataclass
class BenchmarkSeiesStatUnit(SeriesStatUnit):

    alpha: float                    # 超额收益
    beta: float                     # 风险暴露
    track_err: float                # 跟踪误差
    ir: float                       # 信息比率
    alpha_cl: float                 # 择股收益
    beta_cl: float                  # 择时收益
    treynor: float                  # 特雷诺
    win_rate: float                 # 对基准胜率
    win_rate_0: float               # 对零胜率
    continue_value: float           # 业绩持续性
    win_rate_weekly: float          # 周度对基准胜率
    win_rate_0_weekly: float        # 周度对零胜率
    win_rate_monthly: float         # 月度对基准胜率 
    win_rate_0_monthly: float       # 月度对零胜率
    excess_mdd: float               # 超额收益最大回撤
    excess_mdd_date1: datetime.date # 超额收益最大回撤开始日
    excess_mdd_date2: datetime.date # 超额收益最大回撤结束日
    excess_mdd_lens: int            # 超额收益最大回撤持续时间
    up_capture: float               # 向上捕捉率
    down_capture: float             # 向下捕捉率
    not_system_risk: float          # 非系统性风险
    corr: float                     # 相关性
    sortino: float                  # 索提诺
    skew: float                     # 偏度
    kurtosis: float                 # 峰度
    total_relative_ret: float       # 相对全部收益
    asset_to_bmk_mdd:float          # 资产对基准回撤比例

@dataclasses.dataclass
class RetStatUnit:

    recent_natural_year_ret: float      # 今年以来收益 未年化
    recent_1w_ret: float                # 最近一周收益 未年化
    recent_1m_ret: float                # 最近一月收益 未年化
    recent_3m_ret: float                # 最近三月收益 未年化
    recent_6m_ret: float                # 最近半年收益 未年化
    recent_1y_ret: float                # 最近一年收益 未年化
    recent_2y_ret: float                # 最近二年收益 未年化
    recent_3y_ret: float                # 最近三年收益 未年化
    recent_5y_ret: float                # 最近五年收益 未年化
    history_ret: float                  # 成立以来收益 未年化
    history_annual_ret: float           # 成立以来收益 年化

@dataclasses.dataclass
class VolStatUnit:

    recent_natural_year_vol: float      # 今年以来风险 年化
    recent_1w_vol: float                # 最近一周风险 年化
    recent_1m_vol: float                # 最近一月风险 年化
    recent_3m_vol: float                # 最近三月风险 年化
    recent_6m_vol: float                # 最近半年风险 年化
    recent_1y_vol: float                # 最近一年风险 年化
    recent_3y_vol: float                # 最近三年风险 年化
    recent_5y_vol: float                # 最近五年风险 年化
    history_vol: float                  # 成立以来风险 年化

class Calculator:

    TRADING_DAYS_PER_YEAR = 242
    
    @staticmethod
    def get_risk_free_rate():
        return 0.025

    @staticmethod
    def get_yearly_multiplier(frequency:str='daily'):
        # daily weekly 2-weekly monthly yearly
        if frequency == 'daily':
            return 1
        if frequency == 'weekly':
            return 5
        if frequency == '2-weekly':
            return 10
        if frequency == 'monthly':
            return 20
        if frequency == 'yearly':
            return 242

    @staticmethod
    def get_datatime_date(dt):
        dt = dt if isinstance(dt, datetime.datetime) else pandas.to_datetime(dt).date()
        return dt

    @staticmethod
    def get_stat_result(dates: pandas.core.series.Series,
                        values: pandas.core.series.Series,
                        frequency: str='1D',
                        risk_free_rate: float=0.025,
                        ): #['1D','1W','2W','1M']
        res = CalculatorBase.get_stat_result(dates=dates,values=values,frequency=frequency,risk_free_rate=risk_free_rate)
        return SeriesStatUnit(
            start_date=res['start_date'],
            end_date=res['end_date'],
            trade_year=res['trade_year'],
            trade_month=res['trade_month'],
            fund_nav_completion=res['fund_nav_completion'],
            last_unit_nav=res['last_unit_nav'],
            annualized_ret=res['annual_ret'],
            annualized_vol=res['annual_vol'],
            sharpe=res['sharpe'],
            cumu_ret=res['cumu_ret'],
            recent_1w_ret=res['recent_1w_ret'],
            recent_1m_ret=res['recent_1m_ret'],
            recent_4w_ret=res['recent_4w_ret'],
            y_2020_ret=res['y_2020_ret'],
            last_year_ret=res['last_year_ret'],
            recent_3m_ret=res['recent_3m_ret'],
            recent_6m_ret=res['recent_6m_ret'],
            recent_y1_ret=res['recent_1y_ret'],
            recent_y2_ret=res['recent_2y_ret'],
            recent_y3_ret=res['recent_3y_ret'],
            recent_y5_ret=res['recent_5y_ret'],
            worst_3m_ret=res['worst_3m_ret'],
            worst_6m_ret=res['worst_6m_ret'],   
            mdd=res['mdd'],
            mdd_date1=res['mdd_date1'],
            mdd_date2=res['mdd_date2'],
            mdd_lens=res['mdd_lens'],
            ret_over_mdd=res['calmar'],
            last_mv_diff=res['last_mv_diff'],
            last_increase_rate=res['last_increase_rate'],
            recent_drawdown=res['recent_drawdown'],
            recent_mdd_date1 = res['recent_mdd_date1'],
            downside_std=res['downside_risk'],
            var=res['var'],
            cvar=res['cvar'],
            recent_year_ret=res['recent_year_ret'],
            raise_month_num=res['raise_month_num'],
            drop_month_num=res['drop_month_num'],
            mdd_recover=res['mdd_recover'],
            mdd_recover_date1=res['mdd_recover_date1'],
            mdd_recover_date2=res['mdd_recover_date2'],
            mdd_recover_lens=res['mdd_recover_lens'],
            win_rate_0_period=res['win_rate_0_period'],
            nav_density=res['nav_density'],
            nav_type=res['nav_type'],
            w1_begin_date=res['w1_begin_date'],
            w1_end_date=res['w1_end_date'],
            recent_1w_annual_ret=res['recent_1w_annual_ret'],
            recent_1w_mdd=res['recent_1w_mdd'],
            recent_1m_annual_ret=res['recent_1m_annual_ret'],
            recent_1m_mdd=res['recent_1m_mdd'],
            recent_3m_annual_ret=res['recent_3m_annual_ret'],
            recent_3m_mdd=res['recent_3m_mdd'],
            recent_6m_annual_ret=res['recent_6m_annual_ret'],
            recent_6m_mdd=res['recent_6m_mdd'],
            recent_1y_annual_ret=res['recent_1y_annual_ret'],
            recent_1y_mdd=res['recent_1y_mdd'],
            recent_2y_annual_ret=res['recent_2y_annual_ret'],
            recent_2y_mdd=res['recent_2y_mdd'],
            recent_3y_annual_ret=res['recent_3y_annual_ret'],
            recent_3y_mdd=res['recent_3y_mdd'],
            recent_4y_annual_ret=res['recent_4y_annual_ret'],
            recent_4y_mdd=res['recent_4y_mdd'],
            recent_5y_annual_ret=res['recent_5y_annual_ret'],
            recent_5y_mdd=res['recent_5y_mdd'],
            y_last_ret_annual=res['y_last_ret_annual'],
            y_last_mdd=res['y_last_mdd'],
        )

    @staticmethod
    def get_cur_mdd(dates:  pandas.core.series.Series,
                    values: pandas.core.series.Series):
        sr = pandas.Series(values, index=dates).sort_index().dropna()
        df = pandas.DataFrame(sr/sr.cummax() - 1)
        df.columns =['cur_mdd']
        return df

    @staticmethod
    def get_monthly_ret(dates:  pandas.core.series.Series,
                        values: pandas.core.series.Series,
                        ret_period: int=2021):
        df = pandas.Series(values, index=dates).sort_index().dropna()
        df = df.set_axis(pandas.to_datetime(df.index), inplace=False).resample(rule='1M').last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        df = pandas.DataFrame(df.pct_change(1).loc[datetime.date(ret_period,1,1):datetime.date(ret_period+1,1,1)])
        df.columns = ['monthly_ret']
        return df

    @staticmethod
    def get_benchmark_stat_result(dates: pandas.core.series.Series,
                                  values: pandas.core.series.Series,
                                  benchmark_values: pandas.core.series.Series,
                                  frequency:str='1D',
                                  risk_free_rate: float=0.025,
                                  ):#['1D','1W','2W','1M']
        res = CalculatorBase.get_stat_result(dates=dates,values=values,benchmark_values=benchmark_values,frequency=frequency,risk_free_rate=risk_free_rate)
        return BenchmarkSeiesStatUnit(
            start_date=res['start_date'],
            end_date=res['end_date'],
            trade_year=res['trade_year'],
            trade_month=res['trade_month'],
            fund_nav_completion=res['fund_nav_completion'],
            last_unit_nav=res['last_unit_nav'],
            annualized_ret=res['annual_ret'],
            annualized_vol=res['annual_vol'],
            sharpe=res['sharpe'],
            recent_y5_ret=res['recent_5y_ret'],
            mdd=res['mdd'],
            mdd_date1=res['mdd_date1'],
            mdd_date2=res['mdd_date2'],
            mdd_lens=res['mdd_lens'],
            ret_over_mdd=res['calmar'],
            alpha=res['alpha'],
            beta=res['beta'],
            track_err=res['track_err'],
            ir=res['info'],
            last_mv_diff=res['last_mv_diff'],
            last_increase_rate=res['last_increase_rate'],
            recent_1w_ret=res['recent_1w_ret'],
            recent_1m_ret=res['recent_1m_ret'],
            recent_4w_ret=res['recent_4w_ret'],
            y_2020_ret=res['y_2020_ret'],
            last_year_ret=res['last_year_ret'],
            recent_3m_ret=res['recent_3m_ret'],
            recent_6m_ret=res['recent_6m_ret'],
            recent_y1_ret=res['recent_1y_ret'],
            recent_y3_ret=res['recent_3y_ret'],
            worst_3m_ret=res['worst_3m_ret'],
            worst_6m_ret=res['worst_6m_ret'],
            recent_drawdown=res['recent_drawdown'],
            recent_mdd_date1 = res['recent_mdd_date1'],
            downside_std=res['downside_risk'],
            var=res['var'],
            cvar=res['cvar'],
            win_rate=res['win_rate'],
            win_rate_0=res['win_rate_0'],
            treynor=res['treynor'],
            alpha_cl=res['alpha_cl'],
            beta_cl=res['beta_cl'],
            continue_value=res['continue_value_m'],
            recent_year_ret=res['recent_year_ret'],
            raise_month_num=res['raise_month_num'],
            drop_month_num=res['drop_month_num'],
            win_rate_weekly=res['win_rate_weekly'],
            win_rate_0_weekly=res['win_rate_0_weekly'],
            win_rate_monthly=res['win_rate_monthly'],
            win_rate_0_monthly=res['win_rate_0_monthly'],
            excess_mdd=res['excess_mdd'],
            excess_mdd_date1=res['excess_mdd_date1'],
            excess_mdd_date2=res['excess_mdd_date2'],
            excess_mdd_lens=res['excess_mdd_lens'],
            up_capture=res['up_capture'],
            down_capture=res['down_capture'],
            not_system_risk=res['not_system_risk'],
            corr=res['corr'],
            sortino=res['sortino'],
            skew=res['skew'],
            kurtosis=res['kurtosis'],
            cumu_ret=res['cumu_ret'],
            total_relative_ret=res['total_relative_ret'],
            mdd_recover=res['mdd_recover'],
            mdd_recover_date1=res['mdd_recover_date1'],
            mdd_recover_date2=res['mdd_recover_date2'],
            mdd_recover_lens=res['mdd_recover_lens'],
            win_rate_0_period=res['win_rate_0_period'],
            nav_density=res['nav_density'],
            nav_type=res['nav_type'],
            w1_begin_date=res['w1_begin_date'],
            w1_end_date=res['w1_end_date'],
            recent_y2_ret=res['recent_2y_ret'],
            recent_1w_annual_ret=res['recent_1w_annual_ret'],
            recent_1w_mdd=res['recent_1w_mdd'],
            recent_1m_annual_ret=res['recent_1m_annual_ret'],
            recent_1m_mdd=res['recent_1m_mdd'],
            recent_3m_annual_ret=res['recent_3m_annual_ret'],
            recent_3m_mdd=res['recent_3m_mdd'],
            recent_6m_annual_ret=res['recent_6m_annual_ret'],
            recent_6m_mdd=res['recent_6m_mdd'],
            recent_1y_annual_ret=res['recent_1y_annual_ret'],
            recent_1y_mdd=res['recent_1y_mdd'],
            recent_2y_annual_ret=res['recent_2y_annual_ret'],
            recent_2y_mdd=res['recent_2y_mdd'],
            recent_3y_annual_ret=res['recent_3y_annual_ret'],
            recent_3y_mdd=res['recent_3y_mdd'],
            recent_4y_annual_ret=res['recent_4y_annual_ret'],
            recent_4y_mdd=res['recent_4y_mdd'],
            recent_5y_annual_ret=res['recent_5y_annual_ret'],
            recent_5y_mdd=res['recent_5y_mdd'],
            y_last_ret_annual=res['y_last_ret_annual'],
            y_last_mdd=res['y_last_mdd'],
            asset_to_bmk_mdd=res['asset_to_bmk_mdd'], 
        )

    @staticmethod
    def get_recent_ret(dates:             pandas.core.series.Series,
                       values:            pandas.core.series.Series,
                       frequency:         str='1D'):
        res = CalculatorBase.calc_recent_ret(dates=dates,values=values,frequency=frequency)
        return RetStatUnit(
            recent_natural_year_ret=res['recent_natural_year_ret'],
            recent_1w_ret=res['recent_1w_ret'],
            recent_1m_ret=res['recent_1m_ret'],
            recent_3m_ret=res['recent_3m_ret'],
            recent_6m_ret=res['recent_6m_ret'],
            recent_1y_ret=res['recent_1y_ret'],
            recent_2y_ret=res['recent_2y_ret'],
            recent_3y_ret=res['recent_3y_ret'],
            recent_5y_ret=res['recent_5y_ret'],
            history_ret=res['history_ret'],
            history_annual_ret=res['history_annual_ret'],
        )

    @staticmethod
    def get_recent_vol(dates:             pandas.core.series.Series,
                       values:            pandas.core.series.Series,
                       frequency:         str='1D'):
        res = CalculatorBase.calc_recent_vol(dates=dates,values=values,frequency=frequency)
        return VolStatUnit(
            recent_natural_year_vol=res['recent_natural_year_vol'],
            recent_1w_vol=res['recent_1w_vol'],
            recent_1m_vol=res['recent_1m_vol'],
            recent_3m_vol=res['recent_3m_vol'],
            recent_6m_vol=res['recent_6m_vol'],
            recent_1y_vol=res['recent_1y_vol'],
            recent_3y_vol=res['recent_3y_vol'],
            recent_5y_vol=res['recent_5y_vol'],
            history_vol=res['history_vol'],
        )

    @staticmethod
    def get_penetraded_stock_analysis(fund_list: pandas.core.series.Series,
                                      weight_list: pandas.core.series.Series):
        # fund_list 基金代码 比如 ['000017!0', '000020!0', '000031!0', '000051!0', '000061!0'] 形式
        # weight_list 权重代码 比如 [0.1,0.2,0.2,0.2,0.3] 形式
        from ..data.manager.manager_fund import DataManager
        df = pandas.DataFrame({'fund_id':fund_list,'weight':weight_list}).set_index('fund_id')
        fund_info = DataManager.basic_data(func_name=f'get_fund_info', fund_list = fund_list.tolist())
        fund_type = 'stock'
        fund_list = df.index.tolist()
        stock_weights = DataManager.basic_data(func_name=f'get_fund_hold_{fund_type}_latest', fund_list = fund_list)
        stock_weights = stock_weights.set_index('fund_id').drop(columns=['datetime','_update_time'])
        weight_list = [i for i in stock_weights if 'weight' in i]
        stock_weights = stock_weights.replace({None:np.nan})
        for r in stock_weights.iterrows():
            fund_id = r[0]
            fund_weight = df.loc[fund_id, 'weight']
            for weight_i in weight_list:
                stock_weights.loc[fund_id, weight_i] = stock_weights.loc[fund_id, weight_i] / 100 * fund_weight
        _df = stock_weights.join(df)
        _res = []
        for r in stock_weights.iterrows():
            fund_id = r[0]
            for i in range(1,11):
                dic = {
                    'fund_id' : fund_id,
                    f'{fund_type}_name':r[1][f'rank{i}_{fund_type}'],
                    f'{fund_type}_weight':r[1][f'rank{i}_{fund_type}weight'], 
                    f'{fund_type}_id':r[1][f'rank{i}_{fund_type}_code']
                }
                _res.append(dic)
        df = pandas.DataFrame(_res)
        df = pandas.merge(df,fund_info[['fund_id','desc_name']],on='fund_id')
        #_sum = df[f'{fund_type}_weight'].sum()
        df[f'{fund_type}_weight'] = df[f'{fund_type}_weight']# / _sum
        df = df.sort_values(f'{fund_type}_weight',ascending=False)

        stock_list = df.stock_id.tolist()
        stock_info = DataManager.raw_data(func_name=f'get_em_stock_info', stock_list=stock_list)
        industry_info = DataManager.raw_data(func_name='get_em_industry_info',ind_class_type=IndClassType.SWI1)
        stock_info.loc[:,'sw_indus_1'] = stock_info.bl_sws_ind_code.map(lambda x: x.split('-')[0] if x is not None else None)
        _df1 = stock_info[['stock_id','name','sw_indus_1']]
        _df2 = industry_info[['em_id','ind_name']].rename(columns={'em_id':'sw_indus_1'})
        stock_info = pandas.merge(_df1,_df2,on='sw_indus_1')
        df = df.join(stock_info[['stock_id','ind_name']].set_index('stock_id'),on='stock_id')
        stock_price = DataManager.raw_data(func_name='get_em_stock_price_recent',codes=stock_list,columns=['pre_close'])
        stock_price.loc[:,'ret'] =  stock_price.close / stock_price.pre_close - 1
        df.join(stock_price[['stock_id','close','ret']].set_index('stock_id'), on='stock_id')
        return df

    @staticmethod
    def group_left_year(x):
        if x <= 1:
            return '1年'
        if x <= 3:
            return '3年'
        if x <= 5:
            return '5年'
        if x <= 10:
            return '10年'

    @staticmethod
    def get_penetraded_bond_analysis(fund_list: pandas.core.series.Series,
                                      weight_list: pandas.core.series.Series):
        # fund_list 基金代码 比如 ['000005!0', '000015!0', '000016!0', '000032!0', '000033!0']  array 形式
        # weight_list 权重代码 比如 [0.1,0.2,0.2,0.2,0.3] array 形式
        from ..data.manager.manager_fund import DataManager
        df = pandas.DataFrame({'fund_id':fund_list,'weight':weight_list}).set_index('fund_id')
        fund_info = DataManager.basic_data(func_name=f'get_fund_info', fund_list = fund_list.tolist())
        fund_type = 'bond'
        fund_list = df.index.tolist()
        stock_weights = DataManager.basic_data(func_name=f'get_fund_hold_{fund_type}_latest', fund_list = fund_list)
        stock_weights = stock_weights.set_index('fund_id').drop(columns=['datetime','_update_time'])
        weight_list = [i for i in stock_weights if 'weight' in i]
        stock_weights = stock_weights.replace({None:np.nan})
        for r in stock_weights.iterrows():
            fund_id = r[0]
            fund_weight = df.loc[fund_id, 'weight']
            for weight_i in weight_list:
                stock_w = stock_weights.loc[fund_id, weight_i]
                stock_weights.loc[fund_id, weight_i] = stock_w / 100 * fund_weight
        
        _df = stock_weights.join(df)
        _res = []
        for r in stock_weights.iterrows():
            fund_id = r[0]
            for i in range(1,11):
                dic = {
                    'fund_id' : fund_id,
                    f'{fund_type}_name':r[1][f'rank{i}_{fund_type}'],
                    f'{fund_type}_weight':r[1][f'rank{i}_{fund_type}weight'], 
                    f'{fund_type}_id':r[1][f'rank{i}_{fund_type}_code']
                }
                _res.append(dic)
        df = pandas.DataFrame(_res)
        df = pandas.merge(df,fund_info[['fund_id','desc_name']],on='fund_id')
        df = df.dropna(subset=['bond_name'])
        _bond_list = df.bond_id.tolist()
        _bond_rate = DataManager.raw_data(func_name='get_em_bond_rate_maturity',codes=_bond_list)
        df = df.set_index('bond_id').join(_bond_rate[['bond_id','maturity','subject_rate']].set_index('bond_id')).reset_index()
        bond_info = DataManager.raw_data(func_name='get_em_bond_info',bond_id_list=_bond_list)
        _bond_info_bond_list = bond_info.bond_id.tolist()
        _conv_bond_list = [i for i in _bond_list if i not in _bond_info_bond_list]
        bond_info.loc[:,'cbl1_type'] = bond_info.cbl1_type.map(lambda x: x.replace('(中债)','').replace('（中债）',''))
        conv_bond_info = DataManager.raw_data(func_name='get_em_conv_bond_info',bond_id_list=_conv_bond_list)
        if conv_bond_info.empty:
            df = df.join(bond_info[['bond_id','cbl1_type']].set_index('bond_id'), on='bond_id')
        else:
            conv_bond_info = conv_bond_info[['CODES']].rename(columns={'CODES':'bond_id'})
            conv_bond_info.loc[:,'cbl1_type'] = '可转债'
            df = df.join(bond_info[['bond_id','cbl1_type']].append(conv_bond_info[['bond_id','cbl1_type']]).set_index('bond_id'), on='bond_id')
        check_df = df.copy()
        #_sum = df[f'{fund_type}_weight'].sum()
        df[f'{fund_type}_weight'] = df[f'{fund_type}_weight']# / _sum
        df = df.sort_values(f'{fund_type}_weight',ascending=False)
        today = datetime.datetime.now().date()
        td = (df.maturity - today)
        _years = [_.days / 365 for _ in td]
        df.loc[:,'maturity_yearly'] = _years
        df.loc[:,'left_year'] = df.maturity_yearly.map(lambda x: Calculator.group_left_year(x))
        bond_weight_df = df[['cbl1_type','bond_weight']].dropna().groupby('cbl1_type').sum()
        bond_weight_df = bond_weight_df / bond_weight_df.sum()
        bond_rate_df = df[['subject_rate','bond_weight']].dropna().groupby('subject_rate').sum()
        bond_rate_df = bond_rate_df / bond_rate_df.sum()
        left_maturity_year = df[['left_year','bond_weight']].dropna().groupby('left_year').sum()
        left_maturity_year = left_maturity_year / left_maturity_year.sum()
        result = {
            'bond_in_funds': df, #基金组合持有债券穿透分析
            'bond_type_weight': bond_weight_df, #基金组合债券种类分布
            'bond_rate_weight': bond_rate_df, # 基金组合债券评级分布
            'bond_maturity_weight': left_maturity_year, # 债券组合债券到期分布
        }
        return result

    @staticmethod
    def get_fund_last_week_last_trade_df(dt:datetime.date=None, fund_list:list=[]):
        from ..data.manager.manager_fund import DataManager
        # 日期不填默认上周最后交易日开始
        if dt is None:
            dt = DataManager.basic_data(func_name='get_last_week_last_trade_dt')
        fund_nav = DataManager.basic_data(func_name='get_fund_nav_with_date',start_date=dt,fund_list=fund_list)
        fund_nav = fund_nav.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value')
        fund_nav = pandas.DataFrame((fund_nav / fund_nav.iloc[0]).iloc[-1] - 1)
        fund_nav.columns = ['ret']
        return fund_nav

    @staticmethod
    def get_fund_annual_ret_vol_recent_and_history(fund_list:list, hold_date_list:list):
        from ..data.manager.manager_fund import DataManager
        nav = DataManager.basic_data(func_name='get_fund_nav_with_date',fund_list=fund_list)
        nav = nav.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value')
        result = {}
        for fund_id, hold_date in zip(fund_list,hold_date_list):
            result[fund_id] = {}
            for b_d in [None, hold_date]:
                if b_d is None:
                    nav_i = nav[fund_id].dropna()
                else:
                    nav_i = nav[fund_id].loc[b_d:].dropna()
                last_unit_nav = nav_i.iloc[-1] / nav_i.iloc[0]
                days = (nav_i.index[-1] - nav_i.index[0]).days
                annual_ret = math.pow(last_unit_nav, CalculatorBase.TOTAL_DAYS_PER_YEAR/days) - 1
                annual_vol = nav_i.pct_change(1).std(ddof=1) * np.sqrt(CalculatorBase.TRADING_DAYS_PER_YEAR)
                if b_d is None:
                    result[fund_id]['history_annual_ret'] = annual_ret
                    result[fund_id]['history_annual_vol'] = annual_vol
                else:
                    result[fund_id]['hold_annual_ret'] = annual_ret
                    result[fund_id]['hold_annual_vol'] = annual_vol
        return result

    @staticmethod
    def get_manager_result(dates: pandas.core.series.Series,
                          values: pandas.core.series.Series,
                          frequency:str='daily'):
        assert len(dates) == len(values), 'date and value has different numbers'
        
        if len(dates) > 1:
            sr = pandas.Series(values, index=dates).sort_index()
            risk_fee_rate = Calculator.get_risk_free_rate()
            start_date = sr.index[0] 
            end_date = sr.index[-1]
            yearly_multiplier = Calculator.get_yearly_multiplier(frequency)
            trade_year = sr.shape[0] * yearly_multiplier / Calculator.TRADING_DAYS_PER_YEAR
            last_unit_nav = sr[-1] / sr[0]
            annualized_ret = np.exp(np.log(last_unit_nav) / trade_year) - 1
            annualized_vol = (sr / sr.shift(1)).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR / yearly_multiplier)
            sharpe = (annualized_ret - risk_fee_rate) / annualized_vol
            mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
        # 如果只输入一天，返回空
        else:
            start_date = dates[0] if len(dates) > 0 else None
            end_date = dates[0] if len(dates) > 0 else None
            trade_year = 0
            last_unit_nav = 1
            annualized_ret = 0
            annualized_vol = 0
            sharpe = 0 
            mdd = 0

        return {
            'start_date': start_date,
            'end_date': end_date,
            'trade_year' : trade_year,
            'annualized_ret' : annualized_ret,
            'annualized_vol' : annualized_vol,
            'sharpe' : sharpe,
            'mdd' : mdd,
        }
        
    @staticmethod
    def get_manager_stat_result(dates: pandas.core.series.Series,
                                  values: pandas.core.series.Series,
                                  benchmark_values: pandas.core.series.Series,
                                  frequency:str='daily'):
        
        assert len(dates) == len(values), 'date and port_values has different numbers'
        assert len(dates) == len(benchmark_values), 'date and bench_values has different numbers'
        res = Calculator.get_manager_result(dates, values)
        sr_values  = pandas.Series(values, index=dates).sort_index()  
        sr_values = sr_values / sr_values[0]
        yearly_multiplier = Calculator.get_yearly_multiplier(frequency)
        sr_benchmark_values = pandas.Series(benchmark_values, index=dates).sort_index()  
        sr_benchmark_values = sr_benchmark_values / sr_benchmark_values[0]
        sr_values_ret = sr_values.pct_change(1)
        sr_benchmark_ret = sr_benchmark_values.pct_change(1)
        x = sr_benchmark_ret.dropna().values
        y = sr_values_ret.dropna().values
        x = sm.add_constant(x)
        if x.shape[1] < 2:
            return {
                'start_date': res['start_date'],
                'end_date': res['end_date'],
                'trade_year':res['trade_year'],
                'annualized_ret':res['annualized_ret'],
                'annualized_vol':res['annualized_vol'],
                'sharpe':res['sharpe'],
                'mdd':res['mdd'],
                'alpha':np.nan,
                'beta':np.nan,
                'track_err':np.nan,
            }
        model = regression.linear_model.OLS(y,x).fit()
        x = x[:,1]
        alpha = model.params[0] * Calculator.TRADING_DAYS_PER_YEAR / yearly_multiplier
        beta = model.params[1]
        track_err = (sr_values_ret - sr_benchmark_ret).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR / yearly_multiplier)
        
        return {
                'start_date': res['start_date'],
                'end_date': res['end_date'],
                'trade_year':res['trade_year'],
                'annualized_ret':res['annualized_ret'],
                'annualized_vol':res['annualized_vol'],
                'sharpe':res['sharpe'],
                'mdd':res['mdd'],
                'alpha':alpha,
                'beta':beta,
                'track_err':track_err,
        }

    @staticmethod
    def calc_cl_alpha_beta(total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                }
    
    @staticmethod
    def calc_continue_regress_v(monthly_ret):
        # 半年周期
        window = 6
        yearly_multiplier = 2
        period_num = 12
        risk_fee_rate = Calculator.get_risk_free_rate()
        annual_ret = monthly_ret.rolling(window=window).apply(lambda x : x.sum() * yearly_multiplier - risk_fee_rate)
        annual_vol = monthly_ret.rolling(window=window).apply(lambda x : x.std(ddof=1) * np.sqrt(period_num))
        sharpe = (annual_ret / annual_vol)
        sharpe = sharpe.replace(np.Inf, np.nan).replace(-np.Inf, sharpe).dropna() # 空值sharpe赋值np.nan
        if sharpe.shape[0] < 6:
            return np.nan
        mod = AutoReg(endog=sharpe.values, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    @staticmethod
    def get_benchmark(index_price: pandas.DataFrame,
                        weight_list: list,
                        index_list: list,
                        begin_date: datetime.date,
                        end_date: datetime.date)->pandas.DataFrame:

        assert len(weight_list) == len(index_list), 'weight_list and index_list has different numbers'
        assert index_price.shape[1] == len(weight_list), 'columns in index_price are not equal to index list'
        assert begin_date <= end_date, 'end date must bigger than begin date'
        if len(weight_list) < 1:
            return pandas.DataFrame()
        index_price = index_price.loc[begin_date:end_date,].fillna(method='bfill').fillna(method='ffill')
        index_price = index_price / index_price.iloc[0]
        weight_dict = [{'index_id':index_id, 'weight':weight_i } for index_id, weight_i in zip(index_list, weight_list)]
        weight_df = pandas.DataFrame(weight_dict)
        weight_df['weight'] = weight_df['weight'] / weight_df.weight.sum()
        weight_df = weight_df.set_index('index_id').T
        index_price['benchmark'] = index_price.mul(weight_df.values).sum(axis=1)
        return index_price[['benchmark']]

    @staticmethod
    def get_turnover_rate(dates: pandas.core.series.Series,
                          values: pandas.core.series.Series,
                          total_amount: float):
        year_num = len(list(set([_.year for _ in dates])))
        return total_amount / gmean(values) / year_num * 100

    @staticmethod
    def get_stat_result_from_df(df: pandas.DataFrame,
                                date_column: str,
                                value_column: str,
                                frequency: str = '1D'):
        dates = df[date_column].values
        values = df[value_column].values
        return Calculator.get_stat_result(dates, values, frequency)

    @staticmethod
    def get_benchmark_stat_result_from_df(df: pandas.DataFrame,
                                          date_column: str,
                                          port_value_column: str,
                                          bench_value_column: str,
                                          frequency: str = '1D'
                                          ):
        dates = df[date_column].values
        port_values = df[port_value_column].values
        bench_values = df[bench_value_column].values
        return Calculator.get_benchmark_stat_result(dates, port_values, bench_values, frequency)

    @staticmethod
    def get_benchmark_result(index_list:list,
                             weight_list:list,
                             begin_date: datetime.date,
                             end_date: datetime.date) -> pandas.DataFrame:
        '''
        index_list = ['hs300','national_debt']
        weight_list = [0.7,0.3]
        begin_date = datetime.date(2005,1,1)
        end_date = datetime.date(2020,1,1)
        Calculator.get_benchmark_result(index_list, weight_list, begin_date, end_date)
        '''     
        from ..data.api.basic import BasicDataApi              
        basic = BasicDataApi()
        index_price = basic.get_index_price(index_list).pivot_table(index='datetime', columns='index_id', values='close')
        return Calculator.get_benchmark(index_price, weight_list, index_list, begin_date, end_date)
    
    @staticmethod
    def get_index_pct_data_df(index_id:str='hs300',
                           index_val:str=None,
                           windows:int=10*242,
                           min_periods:int=5*242,
                           pct_list:list=[0.15, 0.5, 0.6, 0.8]):
        from ..data.api.derived import DerivedDataApi
        assert (index_id in list(TaaTunerParam.PCT_PAIR.keys())) or (index_val is not None), \
            f'neither index_id{index_id} nor index_val {index_val} is valid'

        index_val = index_val if index_val else TaaTunerParam.PCT_PAIR[index_id]
        # load data
        index_val_data = DerivedDataApi().get_index_valuation_develop_without_date(index_ids=[index_id])
        
        values = index_val_data[index_val].values
        dates = index_val_data.datetime.values
        return  Calculator.get_index_pct_data(dates=dates,
                                            values=values,
                                            pct_list=pct_list,
                                            windows=windows,
                                            min_periods=min_periods)

    @staticmethod
    def roll_pct(x, pct_list, res):
        l = sorted(x)
        lens = len(l)
        res.append([l[int(lens*pct_i)] for pct_i in pct_list])
        return 1

    @staticmethod
    def get_index_pct_data(dates:pandas.core.series.Series,
                              values: pandas.core.series.Series,
                              pct_list:list,
                              windows:int,
                              min_periods:int):
        df = pandas.DataFrame(values, index=dates).dropna()
        res = []
        df[0].rolling(window= windows, min_periods=min_periods).apply(partial(Calculator.roll_pct, pct_list=pct_list, res=res), raw=True)
        return pandas.DataFrame(res, index=dates[-len(res):])

    @staticmethod
    def get_taa_detail(index_id:str='hs300',
                       index_val:str='roe_pct',
                       taa_param:TAAParam=TAAParam(),
                       ):
        
        # load data
        from ..data.api.basic import BasicDataApi
        from ..data.api.derived import DerivedDataApi
        from ..fund.engine.asset_helper import TAAHelper, TAAStatusMode
        _dts = BasicDataApi().get_trading_day_list()
        dts = _dts.datetime 
        index_val_data = DerivedDataApi().get_index_valuation_develop_without_date(index_ids=[index_id])
        index_val_data = index_val_data.set_index('datetime').reindex(dts).fillna(method='ffill')
        asset_func_index = 'hs300'
        index_val_data['index_id'] = asset_func_index
        index_val_data = index_val_data.reset_index().set_index(['datetime','index_id'])
        fake_weight = AssetWeight(**{asset_func_index:1})
        taa_helper = TAAHelper(taa_params=taa_param)
        for dt in dts:
            asset_pct = index_val_data.loc[dt]
            taa_helper.on_price(dt=dt, asset_price=None, cur_saa=fake_weight, asset_pct=asset_pct, select_val={asset_func_index:index_val}, score_dict={})
        df = pandas.DataFrame(taa_helper.tactic_history).T.dropna()
        df['last_date'] = df.index.to_series().shift(1)
        con = df[asset_func_index] != df[asset_func_index].shift(1)
        df_diff_part = df[con].copy()
        df_diff_part = df_diff_part.reset_index().rename(columns={'index':'begin_date'})
        next_date = df_diff_part['last_date'].shift(-1).tolist()
        next_date[-1] = df.index.values[-1]
        df_diff_part['end_date'] = next_date
        df_result = df_diff_part[df_diff_part[asset_func_index] != TAAStatusMode.NORMAL][['begin_date','end_date',asset_func_index]]
        df_result = df_result.rename(columns = {asset_func_index:'status'})
        return df_result.to_dict('records') 

    @staticmethod
    def get_val_level(x):
        val_dict = {
            'PB百分位':'pb_pct',
            'PE百分位':'pe_pct',
            'PS百分位':'ps_pct',
        }
        pct = getattr(x, val_dict[x.tag_method])
        pe = x.pe_ttm
        if None in [pct,pe]:
            return None
        if pct < 0.3 and pe < 20:
            res = 'low'
        elif pct > 0.7 and pe > 20:
            res = 'high'
        else:
            res = 'median'
        return res

    @staticmethod
    def get_index_level(dt:datetime.date=datetime.date(2020,7,20)):
        from ..data.wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
        from ..data.view.basic_models import IndexInfo, IndexValuationLongTerm
        with BasicDatabaseConnector().managed_session() as db_session:
            query = db_session.query(
                    IndexInfo
            )
            index_info = pandas.read_sql(query.statement, query.session.bind)
        with DerivedDatabaseConnector().managed_session() as mn_session:
            query = mn_session.query(
                IndexValuationLongTerm
            ).filter(
                IndexValuationLongTerm.datetime== dt,
            )
            index_val_df = pandas.read_sql(query.statement, query.session.bind)
        index_val = index_val_df.set_index('index_id').join(index_info.set_index('index_id')[['tag_method']]).reset_index()
        index_val['val_level'] = index_val.apply(lambda x : Calculator.get_val_level(x), axis=1 )
        return index_val[['index_id','val_level']]

    @staticmethod
    def get_index_level_index(index_id:str='hs300'):
        from ..data.wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
        from ..data.view.basic_models import IndexInfo, IndexValuationLongTerm
        with BasicDatabaseConnector().managed_session() as db_session:
            query = db_session.query(
                    IndexInfo
            )
            index_info = pandas.read_sql(query.statement, query.session.bind)
        with DerivedDatabaseConnector().managed_session() as mn_session:
            query = mn_session.query(
                IndexValuationLongTerm
            ).filter(
                IndexValuationLongTerm.index_id.in_([index_id]),
            )
            index_val_df = pandas.read_sql(query.statement, query.session.bind)
        index_val = index_val_df.set_index('index_id').join(index_info.set_index('index_id')[['tag_method']]).reset_index()
        index_val['val_level'] = index_val.apply(lambda x : Calculator.get_val_level(x), axis=1 )
        index_val = index_val[['datetime','val_level']]
        res = []
        last_level = 'begin'
        dic = {}
        for i in index_val.itertuples():
            if i.val_level != last_level and not dic:
                dic.update({'start_date':i.datetime})
                dic.update({'val_level':i.val_level})
                last_level = i.val_level

            elif i.val_level != last_level and dic:
                dic.update({'end_date':last_date})
                last_level = i.val_level
                res.append(dic)
                dic = {}
                dic.update({'start_date':i.datetime})
                dic.update({'val_level':i.val_level})
            last_date = i.datetime
        if 'end_date' not in dic:
            dic.update({'end_date':last_date}) 
        res.append(dic)
        res = [i for i in res if i['val_level'] != 'median']
        return res

    @staticmethod
    def get_index_level_index_by_tag_method(index_id: str, tag_method: str):
        from ..data.wrapper.mysql import DerivedDatabaseConnector
        from ..data.view.derived_models import IndexValuationLongTerm
        val_dict = {
            'PB百分位': IndexValuationLongTerm.pb_pct,
            'PE百分位': IndexValuationLongTerm.pe_pct,
            'PS百分位': IndexValuationLongTerm.ps_pct,
        }

        with DerivedDatabaseConnector().managed_session() as mn_session:
            query = mn_session.query(
                IndexValuationLongTerm.pe_ttm,
                IndexValuationLongTerm.datetime,
                val_dict.get(tag_method),
            ).filter(
                IndexValuationLongTerm.index_id.in_([index_id]),
            )
            index_val_df = pandas.read_sql(query.statement, query.session.bind)
        index_val_df['tag_method'] = tag_method
        if len(index_val_df) < 1:
            return []
        index_val_df['val_level'] = index_val_df.apply(lambda x: Calculator.get_val_level(x), axis=1)
        index_val = index_val_df[['datetime', 'val_level']]
        index_val = index_val.dropna(how='any')
        if len(index_val) < 1:
            return []
        res = []
        last_level = 'begin'
        dic = {}
        for i in index_val.itertuples():
            if i.val_level != last_level and not dic:
                dic.update({'start_date': i.datetime})
                dic.update({'val_level': i.val_level})
                last_level = i.val_level

            elif i.val_level != last_level and dic:
                dic.update({'end_date': last_date})
                last_level = i.val_level
                res.append(dic)
                dic = {}
                dic.update({'start_date': i.datetime})
                dic.update({'val_level': i.val_level})
            last_date = i.datetime
        if 'end_date' not in dic:
            dic.update({'end_date': last_date})
        res.append(dic)
        res = [i for i in res if i['val_level'] != 'median']
        return res

    @staticmethod
    def get_mdd_recover_result(mv_df:pandas.DataFrame):

        def _get_mdd_recover_result_df(df:pandas.DataFrame, date_column:str, value_column:str):
            dates = df[date_column].values
            values = df[value_column].values
            return _mdd_recover_analysis(values,dates)

        def _mdd_recover_analysis(values,dates):
            sr = pandas.Series(values, index=dates).sort_index()
            if sr.empty:
                mdd = 0
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
                return mdd, mdd_date1, mdd_date2, mdd_lens
            mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            if mdd == 0:
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
            else:
                mdd_date = mdd_part.idxmin()
                mdd_date1 = sr[:mdd_date].idxmax()
                sr_tmp = sr[mdd_date1:]
                recover_sr = sr_tmp[sr_tmp> sr[mdd_date1]]
                if recover_sr.empty:
                    mdd_date2 = sr_tmp.index[-1]
                else: 
                    mdd_date2 = sr_tmp[sr_tmp> sr[mdd_date1]].index[0]
                mdd_lens = sr.loc[mdd_date1:mdd_date2].shape[0]
            return mdd, mdd_date1, mdd_date2, mdd_lens

        mdd1, mdd_date1_1, mdd_date1_2, mdd_lens1 = _get_mdd_recover_result_df(mv_df.reset_index(),'date','mv')
        mv_df_1 = mv_df.loc[:mdd_date1_1].reset_index()
        mv_df_2 = mv_df.loc[mdd_date1_2:].reset_index()

        mdd2, mdd_date2_1, mdd_date2_2, mdd_lens2 = _get_mdd_recover_result_df(mv_df_1,'date','mv')
        mdd3, mdd_date3_1, mdd_date3_2, mdd_lens3 = _get_mdd_recover_result_df(mv_df_2,'date','mv')
        if mdd3 > mdd2:
            mdd2 = mdd3
            mdd_date2_1 = mdd_date3_1
            mdd_date2_2 = mdd_date3_2
            mdd_lens2 = mdd_lens3

        res = {
            'mdd1':mdd1,
            'mdd1_d1':mdd_date1_1,
            'mdd1_d2':mdd_date1_2,
            'mdd1_lens':mdd_lens1,
            'mdd2':mdd2,
            'mdd2_d1':mdd_date2_1,
            'mdd2_d2':mdd_date2_2,
            'mdd2_len':mdd_lens2
        }

        return res

    @staticmethod
    def get_mdd_result(mv_df:pandas.DataFrame):

        def _get_mdd_result_df(df:pandas.DataFrame, date_column:str, value_column:str):
            dates = df[date_column].values
            values = df[value_column].values
            return _mdd_analysis(values,dates)

        def _mdd_analysis(values,dates):
            sr = pandas.Series(values, index=dates).sort_index()
            if sr.empty:
                mdd = 0
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
                return mdd, mdd_date1, mdd_date2, mdd_lens
            mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            if mdd == 0:
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
            else:
                mdd_date2 = mdd_part.idxmin()
                mdd_date1 = sr[:mdd_date2].idxmax()
                mdd_lens = sr.loc[mdd_date1:mdd_date2].shape[0]
            return mdd, mdd_date1, mdd_date2, mdd_lens

        mdd1, mdd_date1_1, mdd_date1_2, mdd_lens1 = _get_mdd_result_df(mv_df.reset_index(),'date','mv')
        mv_df_1 = mv_df.loc[:mdd_date1_1].reset_index()
        mv_df_2 = mv_df.loc[mdd_date1_2:].reset_index()

        mdd2, mdd_date2_1, mdd_date2_2, mdd_lens2 = _get_mdd_result_df(mv_df_1,'date','mv')
        mdd3, mdd_date3_1, mdd_date3_2, mdd_lens3 = _get_mdd_result_df(mv_df_2,'date','mv')
        if mdd3 > mdd2:
            mdd2 = mdd3
            mdd_date2_1 = mdd_date3_1
            mdd_date2_2 = mdd_date3_2
            mdd_lens2 = mdd_lens3

        res = {
            'mdd1':mdd1,
            'mdd1_d1':mdd_date1_1,
            'mdd1_d2':mdd_date1_2,
            'mdd1_lens':mdd_lens1,
            'mdd2':mdd2,
            'mdd2_d1':mdd_date2_1,
            'mdd2_d2':mdd_date2_2,
            'mdd2_len':mdd_lens2
        }

        return res

    @staticmethod
    def get_asset_stats(df:pandas.DataFrame, asset_info: pandas.DataFrame):
        real_asset_dic = asset_info.set_index('real_name').to_dict()['asset_type']
        indicator_dics = {
            'cumu_ret':'区间收益',
            'annualized_ret':'年化收益',
            'annualized_vol':'年化波动',
            'mdd':'最大回撤',
            'sharpe':'夏普比率',
            'ret_over_mdd':'卡玛比率',
            'downside_std':'下行风险',
        }
        result = []
        for asset in df:
            asset_type = asset_info[asset_info.real_name == asset].asset_type.values[0]
            dates = df[asset].dropna().index
            values = df[asset].dropna().values
            if asset_type != '典型产品':
                dic = Calculator.get_stat_result(dates=dates,values=values).__dict__
            else:
                dic = Calculator.get_stat_result(dates=dates,values=values,frequency='1W').__dict__
            dic['具体指数'] = asset
            dic['类别'] = real_asset_dic[asset]
            result.append(dic)

        df = pandas.DataFrame(result)
        df = df[list(indicator_dics.keys())+['具体指数','类别']].rename(columns=indicator_dics)
        l = ['夏普比率','具体指数','类别','卡玛比率']
        for col in df:
            if col not in l:
                df[col] = (df[col] * 100).round(2)
        df['夏普比率'] = df['夏普比率'].round(2)
        df['卡玛比率'] = df['卡玛比率'].round(2)
        return df
        
    @staticmethod
    def get_index_stats(begin_date, end_date, time_para, index_list):
        try:
            from ..data.api.raw import RawDataApi
            from ..data.api.basic import BasicDataApi
            from ..data.api.derived import DerivedDataApi
            index_info = DerivedDataApi().get_index_val_info()
            res = []
            for i in index_info['index_group']:
                _title = i['title']
                _items = i['items']
                for j in _items:
                    j['title'] = _title
                    res.append(j)
            asset_info = pandas.DataFrame(res).set_index('id')

            indicator_dics = {
                'cumu_ret':'区间收益',
                'annualized_ret':'年化收益',
                'annualized_vol':'年化波动',
                'mdd':'最大回撤',
                'sharpe':'夏普比率',
                'ret_over_mdd':'卡玛比率',
                'downside_std':'下行风险',
            }
            _dic = {
                'pb_mrq':'PB',
                'pe_ttm_nn':'PE',
                'pe_pct_5_3':'PE百分位',
                'roe':'ROE',
                'ps_ttm':'PS',
                'dy':'股息率',
                'eps_ttm':'每股收益',
                'est_peg':'预测PEG',
            }
            result = []
            #asset_info = BasicDataApi().get_asset_info_real(real_id_list=index_list)
            
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            basic = BasicDataApi()
            index_info = basic.get_index_info(index_list=index_list)
            index_name = index_info.set_index('index_id').to_dict()['desc_name']
            df = basic.get_index_price_dt(index_list=index_list,start_date=begin_date,end_date=end_date).pivot_table(index='datetime',columns='index_id',values='close')
            result = []
            for asset in df:
                dates = df[asset].index
                values = df[asset].values
                dic = Calculator.get_stat_result(dates=dates,values=values).__dict__
                dic['具体指数'] = index_name[asset]
                dic['类别'] = asset_info.loc[asset].title
                dic['index_id'] = asset
                result.append(dic)

            df = pandas.DataFrame(result).set_index('index_id')
            df_val = DerivedDataApi().get_index_valuation_last_date(dt=end_date, index_ids=index_list)
            df_val = df_val.set_index('index_id')
            df = df[list(indicator_dics.keys())+['具体指数','类别']].rename(columns=indicator_dics)
            df = df.join(df_val[list(_dic.keys())].rename(columns=_dic))
            l = ['夏普比率','具体指数','卡玛比率','类别']
            for col in df:
                if col not in l and col not in _dic.values():
                    df[col] = (df[col] * 100).round(2)
                if col in _dic.values():
                    df[col] = df[col].round(2)
            df['夏普比率'] = df['夏普比率'].round(2)
            df['卡玛比率'] = df['卡玛比率'].round(2)
            df.loc[:,'EPS增速'] = round(df.PE / df['预测PEG']/100,2)
            return df.reset_index()
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from get_index_stats.__tablename__')
            return pandas.DataFrame([])

    @staticmethod
    def index_val_history(begin_date, end_date, time_para, index_id, valuation_type):
        dic = {
            'pb_mrq':'PB',
            'pe_ttm_nn':'PE',
            'roe':'ROE'
        }
        try:
            from ..data.api.raw import RawDataApi
            from ..data.api.basic import BasicDataApi
            from ..data.api.derived import DerivedDataApi
            derived = DerivedDataApi()
            index_ids = [index_id]
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            index_val = derived.get_index_valuation_develop(index_ids, begin_date, end_date)

            index_val = index_val.rename(columns=dic).set_index('datetime')[[valuation_type]]
            index_val = index_val.replace(0,np.nan).ffill()
            index_price = BasicDataApi().get_index_price_dt(index_list=index_ids,start_date=begin_date,end_date=end_date)

            index_price = index_price.pivot_table(index='datetime',columns='index_id',values='close')
            index_price.columns=['index_price']
            df = index_val.join(index_price).ffill().dropna()
            _df = Calculator.get_index_pct_data(dates=df.index,
                                                values=df[valuation_type].values,
                                                pct_list=[0.1,0.5,0.9],
                                                windows=df.shape[0],
                                                min_periods=60)
            _df.columns = ['10分位','50分位','90分位']
            df = df.join(_df)
            #df.loc[:,'中位数'] = df[valuation_type].rolling(window=df.shape[0], min_periods=10).median()
            #df.loc[:,'百分位'] = df[valuation_type].rolling(window=df.shape[0], min_periods=10).apply(pctrank)
            return df
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.index_val_history')
            return False

    @staticmethod
    def get_product_stats(df:pandas.DataFrame, asset_info: pandas.DataFrame):
        try:
            df = Calculator.get_asset_stats(df, asset_info)
            df = df.rename(columns={'具体指数':'产品'}).drop(columns=['类别'])
            name_dic = {}
            for r in asset_info.itertuples():
                if '典型产品' == r.asset_type:
                    name_dic[r.real_name] = r.asset_name
                else:
                    name_dic[r.real_name] = r.asset_type
            df.loc[:,'策略类型'] = df.产品.map(name_dic)
            return df
        except:
            return pandas.DataFrame()

    @staticmethod
    def get_asset_corr(df:pandas.DataFrame, period='week'):#'month'
        if period == 'week':
            rule = '1W'
        else:
            rule = '1M'
        df = df.set_axis(pandas.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        df = df.pct_change(1).iloc[1:].corr().round(3)
        df = df.sort_values(df.columns[0], ascending=False)
        df.index.name = 'index_id'
        df = df[df.index.tolist()]
        return df

    @staticmethod
    def get_rolling_corr(fund_navs, windows, frequency, step=1):
        try:
            if frequency == '日度':
                rule = '1D'
                if windows <= 2:
                    return pandas.DataFrame()
                windows = int(windows / 1)
            if frequency == '周度':
                rule = '1W'
                if windows <= 10:
                    return pandas.DataFrame()
                windows = int(windows / 5)
            if frequency == '月度':
                rule = '1M'
                if windows <= 40:
                    return pandas.DataFrame()
                windows = int(windows / 20)

            fund_navs_period = CalculatorBase.data_resample_weekly_nav(df=fund_navs, rule=rule)
            fund_ret = fund_navs_period.pct_change(1).iloc[1:]  
            # 弱鸡排列
            col_pair = []
            cols = fund_ret.columns.tolist()
            for idx_i, col_i in enumerate(cols):
                cols_list = cols[idx_i+1:]
                for idx_j, col_j in enumerate(cols_list):
                    col_pair.append([col_i,col_j])

            corr_result_df = []
            for col_pair_i in col_pair:
                df = fund_ret[col_pair_i]
                idx_list = df.reset_index().index.tolist()
                corr_result = []
                b = idx_list[0]
                _result = []
                
                # 计算窗口划分 不包含min_windows概念
                while(1):
                    e = b + windows
                    if e > idx_list[-1]:
                        break
                    else:
                        _result.append([b,e])
                        b += step

                # 计算部分
                for idxs in _result:
                    _df = df.iloc[idxs[0]:idxs[1]]
                    corr_result.append({'datetime':_df.index[-1], f'{col_pair_i[0]}_{col_pair_i[1]}':_df.corr().iloc[1,0]})
                df = pandas.DataFrame(corr_result).set_index('datetime')
                corr_result_df.append(df)
            return pandas.concat(corr_result_df,axis=1).dropna()

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from portfolio.get_rolling_corr')
            return pandas.DataFrame()

    @staticmethod
    def get_product_rolling_corr(fund_navs, windows=20, period='周度', index_id='hs300'):
        try:
            from ..data.api.basic import BasicDataApi
            basic = BasicDataApi()
            try:
                index_name = basic.get_index_info([index_id]).desc_name.tolist()[0]
            except:
                index_name = ''
            if index_name in fund_navs:
                pass
            else:
                try:
                    index_price = basic.get_index_price([index_id]).pivot_table(index='datetime',columns='index_id',values='close')
                    index_price.columns = [index_name]
                    index_price = index_price.reindex(fund_navs.index.union(index_price.index)).ffill()
                    fund_navs = fund_navs.join(index_price).sort_index()
                except:
                    pass

            result = Calculator.get_rolling_corr(fund_navs, windows, period)
            if result.empty:
                df = pandas.DataFrame()
                df.index.name = 'datetime'
                return {
                    'data':df,
                    'details':{},
                }
            if result.shape[0] > 50:
                result = result.iloc[-50:]
            corr_details = {}
            cols = result.columns.tolist()
            for r in result.iterrows():
                _result = []
                for col in cols:
                    x = col.split('_')[0]
                    y = col.split('_')[1]
                    
                    dic_1 = {
                        'x' : x,
                        'y' : y,
                        'v' : r[1][col],
                    }
                    
                    dic_2 = {
                        'x' : y,
                        'y' : x,
                        'v' : r[1][col]
                    }
                    _result.append(dic_1)
                    _result.append(dic_2)
                unique_list = set([item for sublist in cols for item in sublist.split('_')])
                _result += [{'x':i,'y':i,'v':i} for i in  unique_list]
                df_dt = pandas.DataFrame(_result).pivot(index='x',columns='y',values='v')
                df_dt.columns.name = None
                df_dt.index.name = 'index_id'
                corr_details[str(r[0])] = df_dt
            if df_dt.empty:
                df = pandas.DataFrame()
                df.index.name = 'datetime'
                return {
                'data':df,
                'details':corr_details
            }
            return {
                'data':result,
                'details':corr_details
            }

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from calculator.get_product_rolling_corr')
            df = pandas.DataFrame()
            df.index.name = 'datetime'
            return {
                    'data':df,
                    'details':{},
                }

    @staticmethod
    def get_product_rolling_beta(fund_navs, windows=20, period='周度', index_id='hs300'):
        try:
            from ..data.api.basic import BasicDataApi
            if fund_navs.empty:
                df = pandas.DataFrame()
                df.index.name = 'datetime'
                return df
            if period == '日度':
                rule = '1D'
                if windows <= 2:
                    return pandas.DataFrame()
                windows = int(windows / 1)
            if period == '周度':
                rule = '1W'
                if windows <= 10:
                    return pandas.DataFrame()
                windows = int(windows / 5)
            if period == '月度':
                rule = '1M'
                if windows <= 40:
                    return pandas.DataFrame()
                windows = int(windows / 20)

            basic = BasicDataApi()
            index_name = basic.get_index_info([index_id]).desc_name.tolist()[0]
            if index_name in fund_navs:
                pass
            else:
                index_price = basic.get_index_price([index_id]).pivot_table(index='datetime',columns='index_id',values='close')
                index_price.columns = [index_name]
                index_price = index_price.reindex(fund_navs.index.union(index_price.index)).ffill()
                fund_navs = fund_navs.join(index_price).sort_index()

            fund_navs_period = CalculatorBase.data_resample_weekly_nav(df=fund_navs, rule=rule)
            fund_ret = fund_navs_period.pct_change(1).iloc[1:].dropna()  
            result = []
            for fund_name in fund_ret:
                if fund_name == index_name:
                    continue
                else:
                    df = fund_ret[[fund_name, index_name]].rename(columns={fund_name:'fund',index_name:'benchmark'}).reset_index()
                    res = []
                    pandas.Series(df.index).rolling(
                        window=windows).apply(
                        partial(Calculator.rolling_alpha_beta, res=res, df=df), raw=True)
                    df = pandas.DataFrame(res,index=df.datetime[-len(res):])
                    df = df[['beta']].rename(columns={'beta':f'{fund_name}'})
                    result.append(df)
            result = pandas.concat(result, axis=1)
            return result

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from calculator.get_product_rolling_beta')
            return pandas.DataFrame()

    @staticmethod
    def rolling_alpha_beta(x, res, df):
        df_i = df.loc[x[0]:x[-1],]
        return Calculator._rolling_alpha_beta(res, df_i)

    @staticmethod
    def _rolling_alpha_beta(res, df_i):
        if sum(df_i.fund) == 0:
            res.append({'alpha':np.Inf,'beta':np.Inf})
            return 1
        else:
            ploy_res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            p = np.poly1d(ploy_res)
            beta = ploy_res[0]
            alpha = ploy_res[1] * Calculator.TRADING_DAYS_PER_YEAR
            res.append({'alpha':alpha,'beta':beta})
            return 1

    @staticmethod
    def get_product_price_ratio_data(begin_date, end_date, time_para, index_id, fund_id):
        try:
            from ..data.api.raw import RawDataApi
            from ..data.api.basic import BasicDataApi
            raw = RawDataApi()
            basic = BasicDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            hf_info = raw.get_hf_fund_info([fund_id])
            hf_name_dic = hf_info.set_index('fund_id')['desc_name'].to_dict()
            index_info = basic.get_index_info([index_id])
            index_name_dic = index_info.set_index('index_id')['desc_name'].to_dict()

            hf_fund_nav = raw.get_hf_fund_nav(fund_ids=[fund_id],start_date=begin_date,end_date=end_date)
            hf_fund_nav = hf_fund_nav.pivot_table(index='datetime', values='nav', columns='fund_id')
            index_price = basic.get_index_price_dt(index_list=[index_id],start_date=begin_date,end_date=end_date)
            index_price = index_price.pivot_table(index='datetime', values='close', columns='index_id')
            price_df = hf_fund_nav.join(index_price)
            price_df[index_id] = price_df[index_id].ffill()
            price_df = price_df.bfill()
            price_df = price_df/price_df.iloc[0]
            price_df.loc[:,'price_ratio'] = price_df[fund_id] / price_df[index_id]
            hf_name_dic.update(index_name_dic)
            hf_name_dic['price_ratio'] = '价格比'
            indicator_dics = {
                'asset_id':'资产名',
                'cumu_ret':'区间收益',
                'annualized_ret':'年化收益',
                'annualized_vol':'年化波动',
                'mdd':'最大回撤',
                'sharpe':'夏普比率',
                'ret_over_mdd':'卡玛比率',
                'downside_std':'下行风险',
            }
            status_dic = []
            for asset_id in price_df:
                dates = price_df.index
                values = price_df[asset_id]
                dic = Calculator.get_stat_result(dates=dates,values=values,frequency='1W').__dict__
                dic['asset_id'] = asset_id
                status_dic.append(dic)
            df = pandas.DataFrame(status_dic)

            df = df[list(indicator_dics.keys())].rename(columns=indicator_dics)
            l = ['夏普比率','具体指数','资产名','卡玛比率']
            for col in df:
                if col not in l:
                    df[col] = (df[col] * 100).round(2)
            df['夏普比率'] = df['夏普比率'].round(2)
            df['卡玛比率'] = df['卡玛比率'].round(2)
            df['资产名'] = df.资产名.map(hf_name_dic)
            price_df = price_df.rename(columns=hf_name_dic)
            price_df.index.name = 'datetime'
            data = {
                '净值':price_df,
                '指标':df,
            }
            return data

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_product_price_ratio_data')
            return {
                '净值':pandas.DataFrame(),
                '指标':pandas.DataFrame(),
            }

    @staticmethod
    def get_product_price_ratio_data_calcu_part(begin_date, end_date, time_para, index_id, data_dic):
        try:
            from ..data.api.raw import RawDataApi
            from ..data.api.basic import BasicDataApi
            hf_fund_nav = data_dic['hf_fund_nav']
            hf_name_dic = data_dic['hf_name_dic']
            fund_id = hf_fund_nav.columns[0]
            raw = RawDataApi()
            basic = BasicDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            #hf_info = raw.get_hf_fund_info([fund_id])
            #hf_name_dic = hf_info.set_index('fund_id')['desc_name'].to_dict()
            index_info = basic.get_index_info([index_id])
            index_name_dic = index_info.set_index('index_id')['desc_name'].to_dict()

            #hf_fund_nav = raw.get_hf_fund_nav(fund_ids=[fund_id],start_date=begin_date,end_date=end_date)
            #hf_fund_nav = hf_fund_nav.pivot_table(index='datetime', values='nav', columns='fund_id')
            
            index_price = basic.get_index_price_dt(index_list=[index_id],start_date=begin_date,end_date=end_date)
            index_price = index_price.pivot_table(index='datetime', values='close', columns='index_id')
            price_df = hf_fund_nav.join(index_price)
            price_df[index_id] = price_df[index_id].ffill()
            price_df = price_df.bfill()
            price_df = price_df/price_df.iloc[0]
            price_df.loc[:,'price_ratio'] = price_df[fund_id] / price_df[index_id]
            hf_name_dic.update(index_name_dic)
            hf_name_dic['price_ratio'] = '价格比'
            indicator_dics = {
                'asset_id':'资产名',
                'cumu_ret':'区间收益',
                'annualized_ret':'年化收益',
                'annualized_vol':'年化波动',
                'mdd':'最大回撤',
                'sharpe':'夏普比率',
                'ret_over_mdd':'卡玛比率',
                'downside_std':'下行风险',
            }
            status_dic = []
            for asset_id in price_df:
                dates = price_df.index
                values = price_df[asset_id]
                dic = Calculator.get_stat_result(dates=dates,values=values,frequency='1W').__dict__
                dic['asset_id'] = asset_id
                status_dic.append(dic)
            df = pandas.DataFrame(status_dic)

            df = df[list(indicator_dics.keys())].rename(columns=indicator_dics)
            l = ['夏普比率','具体指数','资产名','卡玛比率']
            for col in df:
                if col not in l:
                    df[col] = (df[col] * 100).round(2)
            df['夏普比率'] = df['夏普比率'].round(2)
            df['卡玛比率'] = df['卡玛比率'].round(2)
            df['资产名'] = df.资产名.map(hf_name_dic)
            price_df = price_df.rename(columns=hf_name_dic)
            price_df.index.name = 'datetime'
            data = {
                '净值':price_df,
                '指标':df,
            }
            return data

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_product_price_ratio_data')
            return {
                '净值':pandas.DataFrame(),
                '指标':pandas.DataFrame(),
            }



if __name__ == "__main__":
    # get pct with select rank
    Calculator.get_index_pct_data_df()
    Calculator.get_index_pct_data_df(index_id='hs300',
                           index_val='roe',
                           windows=5*242,
                           min_periods=3*242,
                           pct_list=[0.15, 0.5, 0.6, 0.8])

    # get taa detail
    Calculator.get_taa_detail()
    Calculator.get_taa_detail(index_id='hs300',
                              index_val='pb_pct',
                              taa_param=TAAParam(),
                              )

    Calculator.get_index_level_index(index_id='hs300')